#! /usr/bin/env python
from itertools import zip_longest
from math import pi, ceil, floor, sqrt, atan2, degrees, radians

from solid import union, cube, translate, rotate, square, circle, polyhedron, polygon
from solid import difference, intersection, multmatrix, cylinder, color
from solid import text, linear_extrude, resize
from solid import run_euclid_patch
from solid import OpenSCADObject, P2, P3, P4, Vec3 , Vec4, Vec34, P3s, P23
from solid import Points, Indexes, ScadSize

from euclid3 import Point2, Point3, Vector2, Vector3, Line2, Line3
from euclid3 import LineSegment2, LineSegment3, Matrix4
run_euclid_patch()

# ==========
# = TYPING =
# ==========
from typing import Any, Union, Tuple, Sequence, List, Optional, Callable, Dict, cast
Point23 = Union[Point2, Point3]
Vector23 = Union[Vector2, Vector3]
PointVec23 = Union[Point2, Point3, Vector2, Vector3]
Line23 = Union[Line2, Line3]
LineSegment23 = Union[LineSegment2, LineSegment3]

Tuple2 = Tuple[float, float]
Tuple3 = Tuple[float, float, float]
EucOrTuple = Union[Point3, 
                Vector3, 
                Tuple2, 
                Tuple3
            ]
DirectionLR = float # LEFT or RIGHT in 2D

# =============
# = CONSTANTS =
# =============

EPSILON = 0.01
RIGHT, TOP, LEFT, BOTTOM = range(4)

X, Y, Z = (0, 1, 2)

ORIGIN      = ( 0, 0, 0)
UP_VEC      = ( 0, 0, 1)
RIGHT_VEC   = ( 1, 0, 0)
FORWARD_VEC = ( 0, 1, 0)
DOWN_VEC    = ( 0, 0,-1)
LEFT_VEC    = (-1, 0, 0)
BACK_VEC    = ( 0,-1, 0)

# ==========
# = Colors =
# ========== 
# Deprecated, but kept for backwards compatibility . Note that OpenSCAD natively
# accepts SVG Color names, as seen here: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#color
# From Hans Häggström's materials.scad in MCAD: https://github.com/openscad/MCAD
Red         = (1, 0, 0)
Green       = (0, 1, 0)
Blue        = (0, 0, 1)
Cyan        = (0, 1, 1)
Magenta     = (1, 0, 1)
Yellow      = (1, 1, 0)
Black       = (0, 0, 0)
White       = (1, 1, 1)
Oak         = (0.65, 0.50, 0.40)
Pine        = (0.85, 0.70, 0.45)
Birch       = (0.90, 0.80, 0.60)
FiberBoard  = (0.70, 0.67, 0.60)
BlackPaint  = (0.20, 0.20, 0.20)
Iron        = (0.36, 0.33, 0.33)
Steel       = (0.65, 0.67, 0.72)
Stainless   = (0.45, 0.43, 0.50)
Aluminum    = (0.77, 0.77, 0.80)
Brass       = (0.88, 0.78, 0.50)
Transparent = (1,    1,    1,   0.2)

# ==============
# = Grid Plane =
# ==============
def grid_plane(grid_unit:int=12, count:int=10, line_weight:float=0.1, plane:str='xz') -> OpenSCADObject:

    # Draws a grid of thin lines in the specified plane.  Helpful for
    # reference during debugging.
    l = count * grid_unit
    t = union()
    t.set_modifier('background')
    for i in range(int(-count / 2), int(count / 2 + 1)):
        if 'xz' in plane:
            # xz-plane
            h = up(i * grid_unit)(cube([l, line_weight, line_weight], center=True))
            v = right(i * grid_unit)(cube([line_weight, line_weight, l], center=True))
            t.add([h, v])

        # xy plane
        if 'xy' in plane:
            h = forward(i * grid_unit)(cube([l, line_weight, line_weight], center=True))
            v = right(i * grid_unit)(cube([line_weight, l, line_weight], center=True))
            t.add([h, v])

        # yz plane
        if 'yz' in plane:
            h = up(i * grid_unit)(cube([line_weight, l, line_weight], center=True))
            v = forward(i * grid_unit)(cube([line_weight, line_weight, l], center=True))

            t.add([h, v])

    return t

def distribute_in_grid(objects:Sequence[OpenSCADObject], 
                       max_bounding_box:Tuple[float,float], 
                       rows_and_cols: Tuple[int,int]=None) -> OpenSCADObject:
    # Translate each object in objects in a grid with each cell of size
    # max_bounding_box.
    #
    # objects:  array of SCAD objects
    # max_bounding_box: 2-tuple with x & y dimensions of grid cells.
    #   if a single number is passed, x  & y will both use it
    # rows_and_cols: 2-tuple of how many rows and columns to use. If
    #       not supplied, rows_and_cols will be the smallest square that
    #       can contain all members of objects (e.g, if len(objects) == 80,
    #       rows_and_cols will default to (9,9))

    # Distributes object in a grid in the xy plane
    # with objects spaced max_bounding_box apart
    if isinstance(max_bounding_box, (list, tuple)):
        x_trans, y_trans = max_bounding_box[0:2]
    elif isinstance(max_bounding_box, (int, float, complex)):
        x_trans = y_trans = max_bounding_box
    else:
        pass  # TypeError

    ret = []
    if rows_and_cols:
        grid_h, grid_w = rows_and_cols
    else:
        grid_w = grid_h = int(ceil(sqrt(len(objects))))

    objs_placed = 0
    for y in range(grid_h):
        for x in range(grid_w):
            if objs_placed < len(objects):
                ret.append(
                    translate((x * x_trans, y * y_trans, 0))(objects[objs_placed]))
                objs_placed += 1
            else:
                break
    return union()(*ret)

# ==============
# = Directions =
# ==============
def up(z:float) -> OpenSCADObject:
    return translate((0, 0, z))

def down(z: float) -> OpenSCADObject:
    return translate((0, 0, -z))

def right(x: float) -> OpenSCADObject:
    return translate((x, 0, 0))

def left(x: float) -> OpenSCADObject:
    return translate((-x, 0, 0))

def forward(y: float) -> OpenSCADObject:
    return translate((0, y, 0))

def back(y: float) -> OpenSCADObject:
    return translate((0, -y, 0))

# ===========================
# = Box-alignment rotations =
# ===========================
def rot_z_to_up(obj:OpenSCADObject) -> OpenSCADObject:
    # NOTE: Null op
    return rotate(a=0, v=FORWARD_VEC)(obj)

def rot_z_to_down(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=180, v=FORWARD_VEC)(obj)

def rot_z_to_right(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=90, v=FORWARD_VEC)(obj)

def rot_z_to_left(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=-90, v=FORWARD_VEC)(obj)

def rot_z_to_forward(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=-90, v=RIGHT_VEC)(obj)

def rot_z_to_back(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=90, v=RIGHT_VEC)(obj)

# ================================
# = Box-aligment and translation =
# ================================
def box_align(obj:OpenSCADObject, 
              direction_func:Callable[[float], OpenSCADObject]=up, 
              distance:float=0) -> OpenSCADObject:
    # Given a box side (up, left, etc) and a distance,
    # rotate obj (assumed to be facing up) in the
    # correct direction and move it distance in that
    # direction
    trans_and_rot = {
        up:         rot_z_to_up,  # Null
        down:       rot_z_to_down,
        right:      rot_z_to_right,
        left:       rot_z_to_left,
        forward:    rot_z_to_forward,
        back:       rot_z_to_back,
    }

    assert(direction_func in trans_and_rot)
    rot = trans_and_rot[direction_func]
    return direction_func(distance)(rot(obj))

# =======================
# = 90-degree Rotations =
# =======================

def rot_z_to_x(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=90, v=FORWARD_VEC)(obj)

def rot_z_to_neg_x(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=-90, v=FORWARD_VEC)(obj)

def rot_z_to_neg_y(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=90, v=RIGHT_VEC)(obj)

def rot_z_to_y(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=-90, v=RIGHT_VEC)(obj)

def rot_x_to_y(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=90, v=UP_VEC)(obj)

def rot_x_to_neg_y(obj:OpenSCADObject) -> OpenSCADObject:
    return rotate(a=-90, v=UP_VEC)(obj)

# =======
# = Arc =
# =======
def arc(rad:float, start_degrees:float, end_degrees:float, segments:int=None) -> OpenSCADObject:
    # Note: the circle that this arc is drawn from gets segments,
    # not the arc itself.  That means a quarter-circle arc will
    # have segments/4 segments.

    bottom_half_square = back(rad)(square([3 * rad, 2 * rad], center=True))
    top_half_square = forward(rad)(square([3 * rad, 2 * rad], center=True))

    start_shape = circle(rad, segments=segments)

    if abs((end_degrees - start_degrees) % 360) <= 180:
        end_angle = end_degrees - 180
        ret = difference()(
            start_shape,
            rotate(a=start_degrees)(bottom_half_square.copy()),
            rotate(a=end_angle)(bottom_half_square.copy())
        )
    else:
        ret = intersection()(
            start_shape,
            union()(
                rotate(a=start_degrees)(top_half_square.copy()),
                rotate(a=end_degrees)(bottom_half_square.copy())
            )
        )

    return ret

def arc_inverted(rad:float, start_degrees:float, end_degrees:float, segments:int=None) -> OpenSCADObject:
    # Return the segment of an arc *outside* the circle of radius rad,
    # bounded by two tangents to the circle.  This is the shape
    # needed for fillets.

    # Note: the circle that this arc is drawn from gets segments,
    # not the arc itself.  That means a quarter-circle arc will
    # have segments/4 segments.

    # Leave the portion of a circumscribed square of sides
    # 2*rad that is NOT in the arc behind.  This is most useful for 90-degree
    # segments, since it's what you'll add to create fillets and take away
    # to create rounds.

    # NOTE: an inverted arc is only valid for end_degrees-start_degrees <= 180.
    # If this isn't true, end_degrees and start_degrees will be swapped so
    # that an acute angle can be found.  end_degrees-start_degrees == 180
    # will yield a long rectangle of width 2*radius, since the tangent lines
    # will be parallel and never meet.

    # Fix start/end degrees as needed; find a way to make an acute angle
    if end_degrees < start_degrees:
        end_degrees += 360

    if end_degrees - start_degrees >= 180:
        start_degrees, end_degrees = end_degrees, start_degrees

    # We want the area bounded by:
    # -- the circle from start_degrees to end_degrees
    # -- line tangent to the circle at start_degrees
    # -- line tangent to the circle at end_degrees
    # Note that this shape is only valid if end_degrees - start_degrees < 180,
    # since if the two angles differ by more than 180 degrees,
    # the tangent lines don't converge
    if end_degrees - start_degrees == 180:
        raise ValueError("Unable to draw inverted arc over 180 or more "
                         "degrees. start_degrees: %s end_degrees: %s"
                         % (start_degrees, end_degrees))

    wide = 1000
    high = 1000

    top_half_square = translate((-(wide - rad), 0, 0))(square([wide, high], center=False))
    bottom_half_square = translate((-(wide - rad), -high, 0))(square([wide, high], center=False))

    a = rotate(start_degrees)(top_half_square)
    b = rotate(end_degrees)(bottom_half_square)

    ret = (a * b) - circle(rad, segments=segments)

    return ret

# TODO: arc_to that creates an arc from point to another point.
# This is useful for making paths.  See the SVG path command:
# See: http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes

# ======================
# = Bounding Box Class =
# ======================
class BoundingBox(object):
    # A basic Bounding Box representation to enable some more introspection about
    # objects.  For instance, a BB will let us say "put this new object on top of
    # that old one".  Bounding Boxes *can't* be relied on for boolean operations
    # without compiling in OpenSCAD, so they're limited, but good for some purposes.
    # Be careful to understand what things this BB implementation can and
    # can't do -ETJ 15 Oct 2013

    # Basically you can use a BoundingBox to describe the extents of an object
    # the moment it's created, but once you perform any CSG operation on it, it's
    # more or less useless.
    def __init__(self, size:List[float], loc: List[float]=None):
        loc = loc if loc else [0, 0, 0]
        # self.w, self.h, self.d = size
        # self.x, self.y, self.z = loc
        self.set_size(size)
        self.set_position(loc)

    def size(self) -> List[float]:
        return [self.w, self.h, self.d]

    def position(self) -> List[float]:
        return [self.x, self.y, self.z]

    def set_position(self, position: Sequence[float]):
        self.x, self.y, self.z = position

    def set_size(self, size:Sequence[float]):
        self.w, self.h, self.d = size

    def split_planar(self, 
                     cutting_plane_normal: Vec3=RIGHT_VEC, 
                     cut_proportion: float=0.5, 
                     add_wall_thickness:float=0) -> List['BoundingBox']:
        cpd = {RIGHT_VEC: 0, LEFT_VEC: 0, FORWARD_VEC: 1,
               BACK_VEC: 1, UP_VEC: 2, DOWN_VEC: 2}
        cutting_plane = cpd.get(cutting_plane_normal, 2)

        # Figure what  the cutting plane offset should be
        dim_center = self.position()[cutting_plane]
        dim = self.size()[cutting_plane]
        dim_min = dim_center - dim / 2
        dim_max = dim_center + dim / 2
        cut_point = (cut_proportion) * dim_min + (1 - cut_proportion) * dim_max

        # Now create bounding boxes with the appropriate sizes
        part_bbs = []
        a_sum = 0.0
        for i, part in enumerate([cut_proportion, (1 - cut_proportion)]):
            part_size = self.size()
            part_size[cutting_plane] = part_size[cutting_plane] * part

            part_loc = self.position()
            part_loc[cutting_plane] = dim_min + a_sum + dim * (part / 2)

            # If extra walls are requested around the slices, add them here
            if add_wall_thickness != 0:
                # Expand the walls as requested
                for j in [X, Y, Z]:
                    part_size[j] += 2 * add_wall_thickness
                # Don't expand in the direction of the cutting_plane, only away
                # from it
                part_size[cutting_plane] -= add_wall_thickness

                # add +/- add_wall_thickness/2 to the location in the
                # slicing dimension so we stay at the center of the piece
                loc_offset = -add_wall_thickness / 2 + i * add_wall_thickness
                part_loc[cutting_plane] += loc_offset

            part_bbs.append(BoundingBox(part_size, part_loc))

            a_sum += part * dim

        return part_bbs

    def cube(self, larger: bool=False) -> OpenSCADObject:
        c_size = self.size() if not larger else [s + 2 * EPSILON for s in self.size()]
        c = translate(self.position())(
            cube(c_size, center=True)
        )
        return c

# ===================
# = Model Splitting =
# ===================
def split_body_planar(obj: OpenSCADObject, 
                      obj_bb: BoundingBox, 
                      cutting_plane_normal: Vec3=UP_VEC, 
                      cut_proportion: float=0.5, 
                      dowel_holes: bool=False, 
                      dowel_rad: float=4.5, 
                      hole_depth: float=15, 
                      add_wall_thickness=0) -> Tuple[OpenSCADObject, BoundingBox, OpenSCADObject, BoundingBox]:
    # Split obj along the specified plane, returning two pieces and
    # general bounding boxes for each.
    # Note that the bounding boxes are NOT accurate to the sections,
    # they just indicate which portion of the original BB is in each
    # section.  Given the limits of OpenSCAD, this is the best we can do 
    # -ETJ 17 Oct 2013

    # Optionally, leave holes in both bodies to allow the pieces to be put
    # back together with short dowels.

    # Find the splitting bounding boxes
    part_bbs = obj_bb.split_planar(
        cutting_plane_normal, cut_proportion, add_wall_thickness=add_wall_thickness)

    # And intersect the bounding boxes with the object itself
    slices = [obj * part_bb.cube() for part_bb in part_bbs]

    # Make holes for dowels if requested.
    # In case the bodies need to be aligned properly, make two holes,
    # separated by one dowel-width
    if dowel_holes:
        cpd = {RIGHT_VEC: 0, LEFT_VEC: 0, FORWARD_VEC: 1,
               BACK_VEC: 1, UP_VEC: 2, DOWN_VEC: 2}
        cutting_plane = cpd.get(cutting_plane_normal, 2)

        dowel = cylinder(r=dowel_rad, h=hole_depth * 2, center=True)
        # rotate dowels to correct axis
        if cutting_plane != 2:
            rot_vec = RIGHT_VEC if cutting_plane == 1 else FORWARD_VEC
            dowel = rotate(a=90, v=rot_vec)(dowel)

        cut_point = part_bbs[
            0].position()[cutting_plane] + part_bbs[0].size()[cutting_plane] / 2

        # Move dowels away from center of face by 2*dowel_rad in each
        # appropriate direction
        dowel_trans_a = part_bbs[0].position()
        dowel_trans_a[cutting_plane] = cut_point
        separation_index = {0: 1, 1: 2, 2: 0}[cutting_plane]
        dowel_trans_a[separation_index] -= 2 * dowel_rad
        dowel_trans_b = dowel_trans_a[:]
        dowel_trans_b[separation_index] += 4 * dowel_rad

        dowel_a = translate(dowel_trans_a)(dowel)
        dowel_b = translate(dowel_trans_b)(dowel)

        dowels = dowel_a + dowel_b
        # subtract dowels from each slice
        slices = [s - dowels for s in slices]

    slices_and_bbs = (slices[0], part_bbs[0], slices[1], part_bbs[1])
    return slices_and_bbs

def section_cut_xz(body: OpenSCADObject, y_cut_point:float=0) -> OpenSCADObject:
    big_w = 10000
    d = 2
    c = forward(d / 2 + y_cut_point)(cube([big_w, d, big_w], center=True))
    return c * body

# =====================
# = Bill of Materials =
# =====================
# Any part defined in a method can be automatically counted using the
# `@bom_part()` decorator. After all parts have been created, call
# `bill_of_materials(<SCAD_OBJ>)`
# to generate a report.  See `examples/bom_scad.py` for usage
#
# Additional columns can be added (such as leftover material or URL to part)
# by calling `set_bom_headers()` with a series of string arguments. 
#
# Calling `bom_part()` with additional, non-keyworded arguments will 
# populate the new columns in order of their addition via bom_headers, or 
# keyworded arguments can be used in any order.

g_bom_headers: List[str] = []

def set_bom_headers(*args):
    global g_bom_headers
    g_bom_headers += args

def bom_part(description: str='', per_unit_price:float=None, currency: str='US$', *args, **kwargs) -> Callable:
    def wrap(f):
        name = description if description else f.__name__

        elements = {'name': name, 'Count':0, 'currency':currency, 'Unit Price':per_unit_price}
        # This update also adds empty key value pairs to prevent key exceptions.
        elements.update(dict(zip_longest(g_bom_headers, args, fillvalue='')))
        elements.update(kwargs)

        def wrapped_f(*wargs, **wkwargs):
            scad_obj = f(*wargs, **wkwargs)
            scad_obj.add_trait('BOM', elements)
            return scad_obj

        return wrapped_f

    return wrap

def bill_of_materials(root_obj:OpenSCADObject, csv:bool=False) -> str:
    traits_dicts = _traits_bom_dicts(root_obj)
    # Build a single dictionary from the ones stored on each child object
    # (This is an adaptation of an earlier version, and probably not the most
    # direct way to accomplish this)
    all_bom_traits = {}
    for traits_dict in traits_dicts:
        name = traits_dict['name']
        if name in all_bom_traits:
            all_bom_traits[name]['Count'] += 1
        else:
            all_bom_traits[name] = traits_dict
            all_bom_traits[name]['Count'] = 1
    bom = _make_bom(all_bom_traits, csv)
    return bom

def _traits_bom_dicts(root_obj:OpenSCADObject) -> List[Dict[str, float]]:
    all_child_traits = [_traits_bom_dicts(c) for c in root_obj.children]
    child_traits = [item for subl in all_child_traits for item in subl if item]
    bom_trait = root_obj.get_trait('BOM')
    if bom_trait:
        child_traits.append(bom_trait)
    return child_traits

def _make_bom(bom_parts_dict: Dict[str, float], csv:bool=False, ) -> str:
    field_names = ["Description", "Count", "Unit Price", "Total Price"]
    field_names += g_bom_headers
    
    rows = []
    
    all_costs: Dict[str, float] = {}
    for desc, elements in bom_parts_dict.items():
        row = []
        count = elements['Count']
        currency = elements['currency']
        price = elements['Unit Price']

        if count > 0:
            if price:
                total = price * count
                if currency not in all_costs:
                    all_costs[currency] = 0 
                
                all_costs[currency] += total
                unit_price = _currency_str(price, currency)
                total_price = _currency_str(total, currency)
            else:
                unit_price = total_price = ""
            row = [desc, count, unit_price, total_price]

        for key in g_bom_headers:
            value = elements[key]
            row.append(value)
        rows.append(row)

    # Add total costs if we have values to add
    if len(all_costs) > 0:
        empty_row = [""] * len(field_names)
        rows.append(empty_row)
        for currency, cost in all_costs.items():
            row = empty_row[:]
            row[0] = "Total Cost, {currency:>4}".format(**vars())
            row[3] = "{currency:>4} {cost:.2f}".format(**vars())
            
            rows.append(row)

    res = _table_string(field_names, rows, csv)

    return res

def _currency_str(value:float, currency: str="$") -> str:
    return "{currency:>4} {value:.2f}".format(**vars())
    
def _table_string(field_names: Sequence[str], rows:Sequence[Sequence[float]], csv:bool=False) -> str:
    # Output a justified table string using the prettytable module.
    # Fall back to Excel-ready tab-separated values if prettytable's not found 
    # or CSV is requested
    if not csv:
        try:
            import prettytable
            table = prettytable.PrettyTable(field_names=field_names)
            for row in rows:
                table.add_row(row)
            res = table.get_string()
        except ImportError as e:
            print("Unable to import prettytable module.  Outputting in TSV format")
            csv = True
    if csv:
        lines = ["\t".join(field_names)]
        for row in rows:
            line = "\t".join([str(f) for f in row])
            lines.append(line)

        res = "\n".join(lines) 
        
    return res  + "\n"            

# ================
# = Bounding Box =
# ================
def bounding_box(points: Sequence[EucOrTuple]) -> Tuple[Tuple3, Tuple3]:
    all_x = []
    all_y = []
    all_z = []
    for p in points:
        all_x.append(p[0])
        all_y.append(p[1])
        if len(p) > 2:
            all_z.append(p[2]) # type:ignore 
        else:
            all_z.append(0)

    return ((min(all_x), min(all_y), min(all_z)), (max(all_x), max(all_y), max(all_z)))

# =======================
# = Hardware dimensions =
# =======================
screw_dimensions = {
    'm3': {'nut_thickness': 2.4, 'nut_inner_diam': 5.4, 'nut_outer_diam': 6.1, 'screw_outer_diam': 3.0, 'cap_diam': 5.5, 'cap_height': 3.0},
    'm4': {'nut_thickness': 3.1, 'nut_inner_diam': 7.0, 'nut_outer_diam': 7.9, 'screw_outer_diam': 4.0, 'cap_diam': 6.9, 'cap_height': 3.9},
    'm5': {'nut_thickness': 4.7, 'nut_inner_diam': 7.9, 'nut_outer_diam': 8.8, 'screw_outer_diam': 5.0, 'cap_diam': 8.7, 'cap_height': 5},
}
bearing_dimensions = {
    '608': {'inner_d':8, 'outer_d':22, 'thickness':7},
    '688': {'inner_d':8, 'outer_d':16, 'thickness':5},
    '686': {'inner_d':6, 'outer_d':13, 'thickness':5},
    '626': {'inner_d':6, 'outer_d':19, 'thickness':6},
    '625': {'inner_d':5, 'outer_d':16, 'thickness':5},
    '624': {'inner_d':4, 'outer_d':13, 'thickness':5},
    '623': {'inner_d':3, 'outer_d':10, 'thickness':4},
    '603': {'inner_d':3, 'outer_d':9,  'thickness':5},
    '633': {'inner_d':3, 'outer_d':13, 'thickness':5},
}

def screw(screw_type:str='m3', screw_length:float=16) -> OpenSCADObject:
    dims = screw_dimensions[screw_type.lower()]
    shaft_rad = dims['screw_outer_diam'] / 2
    cap_rad = dims['cap_diam'] / 2
    cap_height = dims['cap_height']

    ret = union()(
        cylinder(shaft_rad, screw_length + EPSILON),
        up(screw_length)(
            cylinder(cap_rad, cap_height)
        )
    )
    return ret

def nut(screw_type:str='m3') -> OpenSCADObject:
    dims = screw_dimensions[screw_type.lower()]
    outer_rad = dims['nut_outer_diam']
    inner_rad = dims['screw_outer_diam']

    ret = difference()(
        circle(outer_rad, segments=6),
        circle(inner_rad)
    )
    return ret

def bearing(bearing_type: str='624') -> OpenSCADObject:
    dims = bearing_dimensions[bearing_type.lower()]
    outerR = dims['outer_d']/2
    innerR = dims['inner_d']/2
    thickness = dims['thickness']
    bearing = cylinder(outerR,thickness)
    bearing.add_param('$fs', 1)
    hole = cylinder(innerR,thickness+2)
    hole.add_param('$fs', 1)
    bearing = difference()(
        bearing,
        translate([0,0,-1])(hole)
    )
    return bearing

# =========
# = LABEL =
# =========
def label(a_str:str, width:float=15, halign:str="left", valign:str="baseline", 
          size:int=10, depth:float=0.5, lineSpacing:float=1.15, 
          font:str="MgOpen Modata:style=Bold", segments:int=40, spacing:int=1) -> OpenSCADObject:
    """Renders a multi-line string into a single 3D object.
    
    __author__    = 'NerdFever.com'
    __copyright__ = 'Copyright 2018-2019 NerdFever.com'
    __version__   = ''
    __email__     = 'dave@nerdfever.com'
    __status__    = 'Development'
    __license__   = Copyright 2018-2019 NerdFever.com
    """

    lines = a_str.splitlines()

    texts = []

    for idx, l in enumerate(lines):
        t = text(text=l, halign=halign, valign=valign, font=font, spacing=spacing).add_param('$fn', segments)
        t = linear_extrude(height=1)(t)
        t = translate([0, -size * idx * lineSpacing, 0])(t)

        texts.append(t)

    result = union()(texts)
    result = resize([width, 0, depth])(result)
    result = translate([0, (len(lines)-1)*size / 2, 0])(result)

    return result

# ==================
# = PyEuclid Utils =
# ==================
def euclidify(an_obj:EucOrTuple, intended_class:type=Vector3) -> Union[Point23, Vector23, List[Union[Point23, Vector23]]]:
    '''
    Accept an object or list of objects of any relevant type (2-tuples, 3-tuples, Vector2/3, Point2/3)
    and return one or more euclid3 objects of intended_class. 

    # -- 3D input has its z-values dropped when intended_class is 2D
    # -- 2D input has its z-values set to 0 when intended_class is 3D

    The general idea is to take in data in whatever form is handy to users
    and return euclid3 types with vector math capabilities
    '''
    sequence = (list, tuple)
    euclidable = (list, tuple, Vector2, Vector3, Point2, Point3)
    numeric = (int, float)
    # If this is a list of lists, return a list of euclid objects
    if isinstance(an_obj, sequence) and isinstance(an_obj[0], euclidable):
        return list((_euc_obj(ao, intended_class) for ao in an_obj))
    elif isinstance(an_obj, euclidable):
        return _euc_obj(an_obj, intended_class)
    else:
        raise TypeError(f'''Object: {an_obj} ought to be PyEuclid class 
                        {intended_class.__name__} or able to form one, but is not.''')
    
def _euc_obj(an_obj: Any, intended_class:type=Vector3) -> Union[Point23, Vector23]:
    ''' Take a single object (not a list of them!) and return a euclid type
        # If given a euclid obj, return the desired type, 
        # -- 3d types are projected to z=0 when intended_class is 2D
        # -- 2D types are projected to z=0 when intended class is 3D
        _euc_obj( Vector3(0,1,2), Vector3) -> Vector3(0,1,2)
        _euc_obj( Vector3(0,1,2), Point3) -> Point3(0,1,2)
        _euc_obj( Vector2(0,1), Vector3) -> Vector3(0,1,0)
        _euc_obj( Vector2(0,1), Point3) -> Point3(0,1,0)
        _euc_obj( (0,1), Vector3) -> Vector3(0,1,0)
        _euc_obj( (0,1), Point3) -> Point3(0,1,0)
        _euc_obj( (0,1), Point2) -> Point2(0,1,0)
        _euc_obj( (0,1,2), Point2) -> Point2(0,1)
        _euc_obj( (0,1,2), Point3) -> Point3(0,1,2)
    '''
    elts_in_constructor = 3
    if intended_class in (Point2, Vector2):
        elts_in_constructor = 2
    result = intended_class(*an_obj[:elts_in_constructor])
    return result

def euc_to_arr(euc_obj_or_list: EucOrTuple) -> List[float]:  # Inverse of euclidify()
    # Call as_arr on euc_obj_or_list or on all its members if it's a list
    result: List[float] = []

    if hasattr(euc_obj_or_list, "as_arr"):
        result = euc_obj_or_list.as_arr()   # type: ignore
    elif isinstance(euc_obj_or_list, (list, tuple)) and hasattr(euc_obj_or_list[0], 'as_arr'):
        result = [euc_to_arr(p) for p in euc_obj_or_list] # type: ignore
    else:
        # euc_obj_or_list is neither an array-based PyEuclid object,
        # nor a list of them.  Assume it's a list of points or vectors,
        # and return the list unchanged.  We could be wrong about this,
        # though.
        result = euc_obj_or_list # type: ignore
    return result

def project_to_2D(euc_obj:Union[Point23, Vector23]) -> Union[Vector2, Point2]:
    """
    Given a Point3/Vector3, return a Point2/Vector2 ignoring the original Z coordinate
    """
    result:Union[Vector2, Point2] = None
    if isinstance(euc_obj, (Point2, Vector2)):
        result = euc_obj
    elif isinstance(euc_obj, Point3):
        result = Point2(euc_obj.x, euc_obj.y)
    elif isinstance(euc_obj, Vector3):
        result = Vector2(euc_obj.x, euc_obj.y)
    else:
        raise ValueError(f"Can't transform object {euc_obj} to a Point2 or Vector2")

    return result

def is_scad(obj:OpenSCADObject) -> bool:
    return isinstance(obj, OpenSCADObject)

def scad_matrix(euclid_matrix4):
    a = euclid_matrix4
    return [[a.a, a.b, a.c, a.d],
            [a.e, a.f, a.g, a.h],
            [a.i, a.j, a.k, a.l],
            [a.m, a.n, a.o, a.p]
            ]

def centroid(points:Sequence[PointVec23]) -> PointVec23:
    if not points:
        raise ValueError(f"centroid(): argument `points` is empty")
    first = points[0]
    is_3d = isinstance(first, (Vector3, Point3))
    if is_3d:
        total = Vector3(0,0,0)
    else:
        total = Vector2(0, 0)

    for p in points:
        total += p
    total /= len(points)

    if isinstance(first, Point2):
        return Point2(*total)
    elif isinstance(first, Point3):
        return Point3(*total)
    else:
        return total

# ==============
# = Transforms =
# ==============
def transform_to_point( body: OpenSCADObject, 
                        dest_point: Point3, 
                        dest_normal: Vector3, 
                        src_point: Point3=Point3(0, 0, 0), 
                        src_normal: Vector3=Vector3(0, 1, 0), 
                        src_up: Vector3=Vector3(0, 0, 1)) -> OpenSCADObject:
    # Transform body to dest_point, looking at dest_normal.
    # Orientation & offset can be changed by supplying the src arguments

    # Body may be:
    #   -- an openSCAD object
    #   -- a list of 3-tuples  or PyEuclid Point3s
    #   -- a single 3-tuple or Point3
    dest_point = euclidify(dest_point, Point3)
    dest_normal = euclidify(dest_normal, Vector3)
    at = dest_point + dest_normal

    EUC_UP = euclidify(UP_VEC)
    EUC_FORWARD = euclidify(FORWARD_VEC)
    EUC_ORIGIN = euclidify(ORIGIN, Vector3)
    # if dest_normal and src_up are parallel, the transform collapses
    # all points to dest_point.  Instead, use EUC_FORWARD if needed
    if dest_normal.cross(src_up) == EUC_ORIGIN:
        if src_up.cross(EUC_UP) == EUC_ORIGIN:
            src_up = EUC_FORWARD
        else:
            src_up = EUC_UP
            
    def _orig_euclid_look_at(eye, at, up):
        '''
        Taken from the original source of PyEuclid's Matrix4.new_look_at() 
        prior to 1184a07d119a62fc40b2c6becdbeaf053a699047 (11 Jan 2015), 
        as discussed here:
        https://github.com/ezag/pyeuclid/commit/1184a07d119a62fc40b2c6becdbeaf053a699047
    
        We were dependent on the old behavior, which is duplicated here:
        '''
        z = (eye - at).normalized()
        x = up.cross(z).normalized()
        y = z.cross(x)

        m = Matrix4.new_rotate_triple_axis(x, y, z)
        m.d, m.h, m.l = eye.x, eye.y, eye.z
        return m
            
    look_at_matrix = _orig_euclid_look_at(eye=dest_point, at=at, up=src_up)
    
    if is_scad(body):
        # If the body being altered is a SCAD object, do the matrix mult
        # in OpenSCAD
        sc_matrix = scad_matrix(look_at_matrix)
        res = multmatrix(m=sc_matrix)(body)
    else:
        body = euclidify(body, Point3)
        if isinstance(body, (list, tuple)):
            res = [look_at_matrix * p for p in body]
        else:
            res = look_at_matrix * body
    return res
    
# ========================================
# = Vector drawing: 3D arrow from a line =
# ========================================
def draw_segment(euc_line: Union[Vector3, Line3]=None, 
                 endless:bool=False, 
                 arrow_rad:float=7, 
                 vec_color: Union[str, Tuple3]=None) -> OpenSCADObject:
    # Draw a traditional arrow-head vector in 3-space.
    vec_arrow_rad = arrow_rad
    vec_arrow_head_rad = vec_arrow_rad * 1.5
    vec_arrow_head_length = vec_arrow_rad * 3

    if isinstance(euc_line, Vector3):
        p = Point3(*ORIGIN)
        v = euc_line
    elif isinstance(euc_line, Line3):
        p = euc_line.p
        v = -euc_line.v
    elif isinstance(euc_line, list) or isinstance(euc_line, tuple):
        # TODO: This assumes p & v are PyEuclid classes.
        # Really, they could as easily be two 3-tuples. Should
        # check for this.
        p, v = euc_line[0], euc_line[1]

    shaft_length = v.magnitude() - vec_arrow_head_length
    arrow = cylinder(r=vec_arrow_rad, h=shaft_length)
    arrow += up(shaft_length)(
        cylinder(r1=vec_arrow_head_rad, r2=0, h=vec_arrow_head_length)
    )
    if endless:
        endless_length = max(v.magnitude() * 10, 200)
        arrow += cylinder(r=vec_arrow_rad / 3,
                            h=endless_length, center=True)

    arrow = transform_to_point(body=arrow, dest_point=p, dest_normal=v)

    if vec_color:
        arrow = color(vec_color)(arrow)

    return arrow

# ==========
# = Offset =
# ==========
# TODO: Make a NamedTuple for LEFT_DIR and RIGHT_DIR
LEFT_DIR, RIGHT_DIR =  1,2

def offset_points(points:Sequence[Point23], 
                  offset:float, 
                  internal:bool=True,
                  closed=True) -> List[Point2]:
    """
    Given a set of points, return a set of points offset by `offset`, in the
    direction specified by `internal`. 

    NOTE: OpenSCAD has the native `offset()` function that generates offset 
    polygons nicely as well as doing fillets & rounds. If you just need a shape,
    prefer using the native `offset()`.  If you need the actual points for some
    purpose, use this function.
    See: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#offset

    # NOTE: We accept Point2s or Point3s, but ignore all Z values and return Point2s

    What is internal or external is defined by by the direction of curvature
    between the first and second points; for non-convex shapes, we will return
    an incorrect (internal points are all external, or vice versa) if the first
    segment pair is concave. This could be mitigated with a point_is_in_polygon()
    function, but I haven't written that yet.
    """
    # Note that we could just call offset_point() repeatedly, but we'd do 
    # a lot of repeated calculations that way
    src_points = euclidify(points, Point2)
    if closed:
        src_points.append(src_points[0])

    vecs = vectors_between_points(src_points)
    direction = direction_of_bend(*src_points[:3])
    if not internal:
        direction = opposite_direction(direction)

    perp_vecs = list((perpendicular_vector(v, direction=direction, length=offset) for v in vecs))

    lines: List[Line2] = []
    for perp, a, b in zip(perp_vecs, src_points[:-1], src_points[1:]):
        lines.append(Line2(a+perp, b+perp))
    
    intersections = list((a.intersect(b) for a,b in zip(lines[:-1], lines[1:])))
    if closed:
        # First point is determined by intersection of first and last lines
        intersections.insert(0, lines[0].intersect(lines[-1]))
    else:
        # otherwise use first and last points in lines
        intersections.insert(0, lines[0].p)
        intersections.append(lines[-1].p + lines[-1].v)
    return intersections

def offset_point(a:Point2, b:Point2, c:Point2, offset:float, direction:DirectionLR=LEFT_DIR) -> Point2:
    ab_perp = perpendicular_vector(b-a, direction, length=offset)
    bc_perp = perpendicular_vector(c-b, direction, length=offset)

    ab_par = Line2(a + ab_perp, b + ab_perp)
    bc_par = Line2(b + bc_perp, c + bc_perp)
    result = ab_par.intersect(bc_par)
    return result

# ==================
# = Offset helpers =
# ==================

def pairwise_zip(l:Sequence) -> zip: # type:ignore
    return zip(l[:-1], l[1:])

def cross_2d(a:Vector2, b:Vector2) -> float:
    """
    scalar value; tells direction of rotation from a to b; 
    see direction_of_bend()
    # See http://www.allenchou.net/2013/07/cross-product-of-2d-vectors/
    """
    return a.x * b.y - a.y * b.x

def direction_of_bend(a:Point2, b:Point2, c:Point2) -> DirectionLR:
    """ 
    Return LEFT_DIR if angle abc is a turn to the left, otherwise RIGHT_DIR
    Returns RIGHT_DIR if ab and bc are colinear
    """
    direction = LEFT_DIR if cross_2d(b-a, c-b) > 0 else RIGHT_DIR
    return direction

def opposite_direction(direction:DirectionLR) -> DirectionLR:
    return LEFT_DIR if direction == RIGHT_DIR else RIGHT_DIR

def perpendicular_vector(v:Vector2, direction:DirectionLR=RIGHT_DIR, length:float=None) -> Vector2:
    perp_vec = Vector2(v.y, -v.x)
    result = perp_vec if direction == RIGHT_DIR else -perp_vec
    if length is not None:
        result.set_length(length)
    return result

def vectors_between_points(points: Sequence[Point23]) -> List[Vector23]:
    """
    Return a list of the vectors from each point in points to the point that follows
    """
    vecs = list((b-a for a,b in pairwise_zip(points))) # type:ignore
    return vecs

# =============
# = 2D Fillet =
# =============

def fillet_2d(three_point_sets: Sequence[Tuple[Point23, Point23, Point23]], 
              orig_poly: OpenSCADObject, 
              fillet_rad: float, 
              remove_material: bool=True) -> OpenSCADObject:
    """
    Return a polygon with arcs of radius `fillet_rad` added/removed (according to
    `remove_material`) to corners specified in `three_point_sets`. 

    e.g. Turn a sharp external corner to a rounded one, or add material
    to a sharp interior corner to smooth it out.
    """
    arc_objs: List[OpenSCADObject] = []
    # TODO: accept Point3s, and project them all to z==0
    for three_points in three_point_sets:
        a, b, c = (project_to_2D(p) for p in three_points)
        ab = a - b
        bc = b - c

        direction = direction_of_bend(a, b, c)

        # center lies at the intersection of two lines parallel to
        # ab and bc, respectively, each offset from their respective 
        # line by fillet_rad
        ab_perp = perpendicular_vector(ab, direction, length=fillet_rad)
        bc_perp = perpendicular_vector(bc, direction, length=fillet_rad)
        center = offset_point(a,b,c, offset=fillet_rad, direction=direction)
        # start_pt = center + ab_perp
        # end_pt = center + bc_perp

        start_degrees = degrees(atan2(ab_perp.y, ab_perp.x))
        end_degrees   = degrees(atan2(bc_perp.y, bc_perp.x))

        # Widen start_degrees and end_degrees slightly so they overlap the areas
        # they're supposed to join/ remove.
        start_degrees, end_degrees = _widen_angle_for_fillet(start_degrees, end_degrees)

        arc_obj = translate(center.as_arr())(
            arc_inverted(rad=fillet_rad, start_degrees=start_degrees, end_degrees=end_degrees)
        )
        arc_objs.append(arc_obj)

    if remove_material:
        poly = orig_poly - arc_objs 
    else:
        poly = orig_poly + arc_objs 

    return poly 

def _widen_angle_for_fillet(start_degrees:float, end_degrees:float) -> Tuple[float, float]:
    # Fix start/end degrees as needed; find a way to make an acute angle
    if end_degrees < start_degrees:
        end_degrees += 360

    if end_degrees - start_degrees >= 180:
        start_degrees, end_degrees = end_degrees, start_degrees

    epsilon_degrees = 0.1
    return start_degrees - epsilon_degrees, end_degrees + epsilon_degrees

# ==============
# = 2D DRAWING =
# ==============
def path_2d(points:Sequence[Point23], width:float=1, closed:bool=False) -> List[Point2]:
    '''
    Return a set of points describing a path of width `width` around `points`,
    suitable for use as a polygon().  

    Note that if `closed` is True, the polygon will have a hole in it, meaning
    that `polygon()` would need to specify its `paths` argument. Assuming 3 elements
    in the original `points` list, we'd have to call:
    path_points = path_2d(points, closed=True)
    poly = polygon(path_points, paths=[[0,1,2],[3,4,5]])

    Or, you know, just call `path_2d_polygon()` and let it do that for you
    '''
    p_a = offset_points(points, offset=width/2, internal=True, closed=closed)
    p_b = list(reversed(offset_points(points, offset=width/2, internal=False, closed=closed)))
    return p_a + p_b

def path_2d_polygon(points:Sequence[Point23], width:float=1, closed:bool=False) -> polygon:
    '''
    Return an OpenSCAD `polygon()` in an area `width` units wide around `points`
    '''
    path_points = path_2d(points, width, closed)
    paths = [list(range(len(path_points)))]
    if closed:
        paths = [list(range(len(points))), list(range(len(points), len(path_points)))]
    return polygon(path_points, paths=paths)

# =================
# = NUMERIC UTILS =
# =================
def frange(start:float, end:float, num_steps:int=None, step_size:float=1.0, include_end=True):
    # if both step_size AND num_steps are supplied, num_steps will be used
    step_size = step_size or 1.0

    if num_steps:
        step_count = num_steps - 1 if include_end else num_steps
        step_size = (end - start)/step_count
    mode = 3 if include_end else 1
    return _frange_orig(start, end, step_size, mode)

def _frange_orig(*args):
    """
    # {{{ http://code.activestate.com/recipes/577068/ (r1)
    frange([start, ] end [, step [, mode]]) -> generator

    A float range generator. If not specified, the default start is 0.0
    and the default step is 1.0.

    Optional argument mode sets whether frange outputs an open or closed
    interval. mode must be an int. Bit zero of mode controls whether start is
    included (on) or excluded (off); bit one does the same for end. Hence:

        0 -> open interval (start and end both excluded)
        1 -> half-open (start included, end excluded)
        2 -> half open (start excluded, end included)
        3 -> closed (start and end both included)

    By default, mode=1 and only start is included in the output.
    """
    mode = 1  # Default mode is half-open.
    n = len(args)
    if n == 1:
        args = (0.0, args[0], 1.0)
    elif n == 2:
        args = args + (1.0,)
    elif n == 4:
        mode = args[3]
        args = args[0:3]
    elif n != 3:
        raise TypeError('frange expects 1-4 arguments, got %d' % n)
    assert len(args) == 3
    try:
        start, end, step = [a + 0.0 for a in args]
    except TypeError:
        raise TypeError('arguments must be numbers')
    if step == 0.0:
        raise ValueError('step must not be zero')
    if not isinstance(mode, int):
        raise TypeError('mode must be an int')
    if mode & 1:
        i, x = 0, start
    else:
        i, x = 1, start + step
    if step > 0:
        if mode & 2:
            from operator import le as comp
        else:
            from operator import lt as comp
    else:
        if mode & 2:
            from operator import ge as comp
        else:
            from operator import gt as comp
    while comp(x, end):
        yield x
        i += 1
        x = start + i * step

def clamp(val: float, min_val: float, max_val: float) -> float:
    result = max(min(val, max_val), min_val)
    return result

def lerp(val: float, min_in: float, max_in: float, min_out: float, max_out: float)-> float:
    if min_in == max_in or min_out == max_out:
        return min_out

    ratio = (val - min_in) / (max_in - min_in)
    result = min_out + ratio * (max_out - min_out);
    return result;

# =====================
# = D e b u g g i n g =
# =====================
def obj_tree_str(sp_obj:OpenSCADObject, vars_to_print:Sequence[str]=None) -> str:
    # For debugging.  This prints a string of all of an object's
    # children, with whatever attributes are specified in vars_to_print

    # Takes an optional list (vars_to_print) of variable names to include in each
    # element (e.g. ['is_part_root', 'is_hole', 'name'])
    if not vars_to_print:
        vars_to_print = []

    # Signify if object has parent or not
    parent_sign = "\nL " if sp_obj.parent else "\n* "

    # Print object
    s = parent_sign + str(sp_obj) + "\t"

    # Extra desired fields
    for v in vars_to_print:
        if hasattr(sp_obj, v):
            s += "%s: %s\t" % (v, getattr(sp_obj, v))

    # Add all children
    for c in sp_obj.children:
        s += indent(obj_tree_str(c, vars_to_print)) # type: ignore

    return s

# =====================
# = DEPENDENT IMPORTS =
# =====================
# imported here to mitigate import loops
from solid.extrude_along_path import extrude_along_path



from ply import lex, yacc

#workaround relative imports.... make this module runable as script
if __name__ == "__main__":
    from scad_ast import *
    from scad_tokens import *
else:
    from .scad_ast import *
    from .scad_tokens import *

precedence = (
    ('nonassoc', 'ASSERT'),
    ('nonassoc', 'ECHO'),
    ('nonassoc', "THEN"),
    ('nonassoc', "ELSE"),
    ('nonassoc', "?"),
    ('nonassoc', ":"),
    ('nonassoc', "(", ")", "{", "}"),

    ('nonassoc', '='),
    ('left', "AND", "OR"),
    ('left', "EQUAL", "NOT_EQUAL", "GREATER_OR_EQUAL", "LESS_OR_EQUAL", ">", "<"),
    ('left', '+', '-'),
    ('left', "%"),
    ('left', '*', '/'),
    ('right', '^'),
    ('right', 'NEG', 'POS', 'BACKGROUND', 'NOT'),
    ('left', "ACCESS"),
 )

def p_statements(p):
    '''statements : statements statement'''
    p[0] = p[1]
    if p[2] != None:
        p[0].append(p[2])

def p_statements_empty(p):
    '''statements : empty'''
    p[0] = []

def p_empty(p):
    'empty : '

def p_statement(p):
    ''' statement : IF "(" expression ")" statement %prec THEN
                |   IF "(" expression ")" statement ELSE statement
                |   for_loop statement
                |   LET "(" assignment_list ")" statement %prec THEN
                |   ASSERT "(" opt_call_parameter_list ")" statement
                |   ECHO "(" opt_call_parameter_list ")" statement
                |   "{" statements "}"
                |   "%" statement %prec BACKGROUND
                |   "*" statement %prec BACKGROUND
                |   "!" statement %prec BACKGROUND
                |   "#" statement %prec BACKGROUND
                |   call statement
                |   USE FILENAME
                |   INCLUDE FILENAME
                |   ";"
                '''
def p_for_loop(p):
    '''for_loop : FOR "(" parameter_list ")"
                | FOR "(" parameter_list ";" expression ";" parameter_list ")"'''

def p_statement_function(p):
    'statement : function'
    p[0] = p[1]

def p_statement_module(p):
    'statement : module'
    p[0] = p[1]

def p_statement_assignment(p):
    'statement : ID "=" expression ";"'
    p[0] = ScadGlobalVar(p[1])

def p_logic_expr(p):
    '''logic_expr :  "-" expression %prec NEG
                |   "+" expression %prec POS
                |   "!" expression %prec NOT
                |   expression "?" expression ":" expression
                |   expression "%" expression
                |   expression "+" expression
                |   expression "-" expression
                |   expression "/" expression
                |   expression "*" expression
                |   expression "^" expression
                |   expression "<" expression
                |   expression ">" expression
                |   expression EQUAL expression
                |   expression NOT_EQUAL expression
                |   expression GREATER_OR_EQUAL expression
                |   expression LESS_OR_EQUAL expression
                |   expression AND expression
                |   expression OR expression
       '''

def p_access_expr(p):
    '''access_expr : ID %prec ACCESS
                |   expression "." ID %prec ACCESS
                |   expression "(" call_parameter_list ")" %prec ACCESS
                |   expression "(" ")" %prec ACCESS
                |   expression "[" expression "]" %prec ACCESS
        '''

def p_list_stuff(p):
    '''list_stuff : FUNCTION "(" opt_parameter_list ")" expression
                |   LET "(" assignment_list ")" expression %prec THEN
                |   EACH expression %prec THEN
                |   "[" expression ":" expression "]"
                |   "[" expression ":" expression ":" expression "]"
                |   "[" for_loop expression "]"
                |   tuple
        '''

def p_assert_or_echo(p):
    '''assert_or_echo : ASSERT "(" opt_call_parameter_list ")"
                    |   ECHO "(" opt_call_parameter_list ")"
       '''
def p_constants(p):
    '''constants : STRING
                |  TRUE
                |  FALSE
                |  NUMBER'''

def p_opt_else(p):
    '''opt_else :
                | ELSE expression %prec THEN
       '''
       #this causes some shift/reduce conflicts, but I don't know how to solve it
def p_for_or_if(p):
    '''for_or_if :  for_loop expression %prec THEN
                |   IF "(" expression ")" expression opt_else
       '''

def p_expression(p):
    '''expression : access_expr
                |   logic_expr
                |   list_stuff
                |   assert_or_echo
                |   assert_or_echo expression %prec ASSERT
                |   constants
                |   for_or_if
                |   "(" expression ")"
       '''
       #the assert_or_echo stuff causes some shift/reduce conflicts, but I don't know how to solve it

def p_assignment_list(p):
    '''assignment_list : ID "=" expression
                    |   assignment_list "," ID "=" expression
       '''

def p_call(p):
    ''' call : ID "(" call_parameter_list ")"
            |  ID "(" ")"'''

def p_tuple(p):
    ''' tuple : "[" opt_expression_list "]"
        '''

def p_commas(p):
    '''commas : commas ","
              | ","
       '''

def p_opt_expression_list(p):
    '''opt_expression_list : expression_list
                        |    expression_list commas
                        |    empty'''
def p_expression_list(p):
    ''' expression_list : expression_list commas expression
                    |     expression
        '''

def p_opt_call_parameter_list(p):
    '''opt_call_parameter_list :
                               | call_parameter_list
       '''
def p_call_parameter_list(p):
    '''call_parameter_list : call_parameter_list commas call_parameter
                        |    call_parameter'''

def p_call_parameter(p):
    '''call_parameter : expression
                    |   ID "=" expression'''

def p_opt_parameter_list(p):
    '''opt_parameter_list : parameter_list
                        |   parameter_list commas
                        |   empty
       '''
    if p[1] != None:
        p[0] = p[1]
    else:
       p[0] = []

def p_parameter_list(p):
    '''parameter_list :     parameter_list commas parameter
                        |   parameter'''
    if len(p) > 2:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_parameter(p):
    '''parameter : ID
                |  ID "=" expression'''
    p[0] = ScadParameter(p[1], len(p) == 4)

def p_function(p):
    '''function : FUNCTION ID "(" opt_parameter_list ")" "=" expression
       '''
    params = None
    if p[4] != ")":
        params = p[4]

    p[0] = ScadFunction(p[2], params)

def p_module(p):
    '''module : MODULE ID "(" opt_parameter_list ")" statement
       '''
    params = None
    if p[4] != ")":
        params = p[4]

    p[0] = ScadModule(p[2], params)


def p_error(p):
    print(f'py_scadparser: Syntax error: {p.lexer.filename}({p.lineno}) {p.type} - {p.value}')

def parseFile(scadFile):

    lexer = lex.lex(debug=False)
    lexer.filename = scadFile
    parser = yacc.yacc(debug=False)

    modules = []
    functions = []
    globalVars = []

    appendObject = { ScadTypes.MODULE : lambda x: modules.append(x),
                     ScadTypes.FUNCTION: lambda x: functions.append(x),
                     ScadTypes.GLOBAL_VAR: lambda x: globalVars.append(x),
    }

    from pathlib import Path
    with Path(scadFile).open() as f:
        for i in  parser.parse(f.read(), lexer=lexer):
            appendObject[i.getType()](i)

    return modules, functions, globalVars

def parseFileAndPrintGlobals(scadFile):

    print(f'======{scadFile}======')
    modules, functions, globalVars = parseFile(scadFile)

    print("Modules:")
    for m in modules:
        print(f'    {m}')

    print("Functions:")
    for m in functions:
        print(f'    {m}')

    print("Global Variables:")
    for m in globalVars:
        print(f'    {m.name}')

if __name__ == "__main__":
    import sys

    if len(sys.argv) < 2:
        print(f"usage: {sys.argv[0]} [-q] <scad-file> [<scad-file> ...]\n   -q : quiete")

    quiete = sys.argv[1] == "-q"
    files = sys.argv[2:] if quiete else sys.argv[1:]

    for i in files:
        if quiete:
            print(i)
            parseFile(i)
        else:
            parseFileAndPrintGlobals(i)




"""
Classes for OpenSCAD builtins
"""
from pathlib import Path
from types import SimpleNamespace
from typing import Dict, Optional, Sequence, Tuple, Union, List

from .solidpython import IncludedOpenSCADObject, OpenSCADObject

PathStr = Union[Path, str]

P2 = Tuple[float, float]
P3 = Tuple[float, float, float]
P4 = Tuple[float, float, float, float]
Vec3 = P3
Vec4 = P4
Vec34 = Union[Vec3, Vec4]
P3s = Sequence[P3]
P23 = Union[P2, P3]
Points = Sequence[P23]
Indexes = Union[Sequence[int], Sequence[Sequence[int]]]
ScadSize = Union[int, Sequence[float]]
OpenSCADObjectPlus = Union[OpenSCADObject, Sequence[OpenSCADObject]]

IMPORTED_SCAD_MODULES: Dict[Path, SimpleNamespace] = {}

class polygon(OpenSCADObject):
    """
    Create a polygon with the specified points and paths.

    :param points: the list of points of the polygon
    :type points: sequence of 2 element sequences

    :param paths: Either a single vector, enumerating the point list, ie. the 
    order to traverse the points, or, a vector of vectors, ie a list of point 
    lists for each separate curve of the polygon. The latter is required if the 
    polygon has holes. The parameter is optional and if omitted the points are 
    assumed in order. (The 'pN' components of the *paths* vector are 0-indexed 
    references to the elements of the *points* vector.)

    :param convexity: OpenSCAD's convexity... yadda yadda

    NOTE: OpenSCAD accepts only 2D points for `polygon()`. Convert any 3D points
    to 2D before compiling
    """

    def __init__(self, points: Union[Points, IncludedOpenSCADObject], paths: Indexes = None, convexity: int = None) -> None:
        # Force points to 2D if they're defined in Python, pass through if they're
        # included OpenSCAD code
        pts = points # type: ignore
        if not isinstance(points, IncludedOpenSCADObject):
            pts = list([(p[0], p[1]) for p in points]) # type: ignore

        args = {'points':pts, 'convexity':convexity}
        # If not supplied, OpenSCAD assumes all points in order for paths
        if paths:
            args['paths'] = paths # type: ignore
        super().__init__('polygon', args)


class circle(OpenSCADObject):
    """
    Creates a circle at the origin of the coordinate system. The argument
    name is optional.

    :param r: This is the radius of the circle. Default value is 1.
    :type r: number

    :param d: This is the diameter of the circle. Default value is 1.
    :type d: number

    :param segments: Number of fragments in 360 degrees.
    :type segments: int
    """

    def __init__(self, r: float = None, d: float = None, segments: int = None) -> None:
        super().__init__('circle',
                         {'r': r, 'd': d, 'segments': segments})


class square(OpenSCADObject):
    """
    Creates a square at the origin of the coordinate system. When center is
    True the square will be centered on the origin, otherwise it is created
    in the first quadrant. The argument names are optional if the arguments
    are given in the same order as specified in the parameters

    :param size: If a single number is given, the result will be a square with 
    sides of that length. If a 2 value sequence is given, then the values will 
    correspond to the lengths of the X and Y sides.  Default value is 1.
    :type size: number or 2 value sequence

    :param center: This determines the positioning of the object. If True, 
    object is centered at (0,0). Otherwise, the square is placed in the positive 
    quadrant with one corner at (0,0). Defaults to False.
    :type center: boolean
    """

    def __init__(self, size: ScadSize = None, center: bool = None) -> None:
        super().__init__('square',
                         {'size': size, 'center': center})


class sphere(OpenSCADObject):
    """
    Creates a sphere at the origin of the coordinate system. The argument
    name is optional.

    :param r: Radius of the sphere.
    :type r: number

    :param d: Diameter of the sphere.
    :type d: number

    :param segments: Resolution of the sphere
    :type segments: int
    """

    def __init__(self, r: float = None, d: float = None, segments: int = None) -> None:
        super().__init__('sphere',
                         {'r': r, 'd': d, 'segments': segments})


class cube(OpenSCADObject):
    """
    Creates a cube at the origin of the coordinate system. When center is
    True the cube will be centered on the origin, otherwise it is created in
    the first octant. The argument names are optional if the arguments are
    given in the same order as specified in the parameters

    :param size: If a single number is given, the result will be a cube with 
    sides of that length. If a 3 value sequence is given, then the values will 
    correspond to the lengths of the X, Y, and Z sides. Default value is 1.
    :type size: number or 3 value sequence

    :param center: This determines the positioning of the object. If True, 
    object is centered at (0,0,0). Otherwise, the cube is placed in the positive 
    quadrant with one corner at (0,0,0). Defaults to False
    :type center: boolean
    """

    def __init__(self, size: ScadSize = None, center: bool = None) -> None:
        super().__init__('cube',
                         {'size': size, 'center': center})


class cylinder(OpenSCADObject):
    """
    Creates a cylinder or cone at the origin of the coordinate system. A
    single radius (r) makes a cylinder, two different radii (r1, r2) make a
    cone.

    :param h: This is the height of the cylinder. Default value is 1.
    :type h: number

    :param r: The radius of both top and bottom ends of the cylinder. Use this 
    parameter if you want plain cylinder. Default value is 1.
    :type r: number

    :param r1: This is the radius of the cone on bottom end. Default value is 1.
    :type r1: number

    :param r2: This is the radius of the cone on top end. Default value is 1.
    :type r2: number

    :param d: The diameter of both top and bottom ends of the cylinder.  Use t
    his parameter if you want plain cylinder. Default value is 1.
    :type d: number

    :param d1: This is the diameter of the cone on bottom end. Default value is 1.
    :type d1: number

    :param d2: This is the diameter of the cone on top end. Default value is 1.
    :type d2: number

    :param center: If True will center the height of the cone/cylinder around 
    the origin. Default is False, placing the base of the cylinder or r1 radius 
    of cone at the origin.
    :type center: boolean

    :param segments: Number of fragments in 360 degrees.
    :type segments: int
    """

    def __init__(self, r: float = None, h: float = None, r1: float = None, r2: float = None,
                 d: float = None, d1: float = None, d2: float = None, center: bool = None,
                 segments: int = None) -> None:
        super().__init__('cylinder',
                         {'r': r, 'h': h, 'r1': r1, 'r2': r2, 'd': d,
                          'd1': d1, 'd2': d2, 'center': center,
                          'segments': segments})


class polyhedron(OpenSCADObject):
    """
    Create a polyhedron with a list of points and a list of faces. The point
    list is all the vertices of the shape, the faces list is how the points
    relate to the surfaces of the polyhedron.

    *note: if your version of OpenSCAD is lower than 2014.03 replace "faces"
    with "triangles" in the below examples*

    :param points: sequence of points or vertices (each a 3 number sequence).

    :param triangles: (*deprecated in version 2014.03, use faces*) vector of 
    point triplets (each a 3 number sequence). Each number is the 0-indexed point 
    number from the point vector.

    :param faces: (*introduced in version 2014.03*) vector of point n-tuples 
    with n >= 3. Each number is the 0-indexed point number from the point vector.  
    That is, faces=[[0,1,4]] specifies a triangle made from the first, second, 
    and fifth point listed in points. When referencing more than 3 points in a 
    single tuple, the points must all be on the same plane.

    :param convexity: The convexity parameter specifies the maximum number of 
    front sides (back sides) a ray intersecting the object might penetrate. This 
    parameter is only needed for correctly displaying the object in OpenCSG 
    preview mode and has no effect on the polyhedron rendering.
    :type convexity: int
    """

    def __init__(self, points: P3s, faces: Indexes, convexity: int = 10, triangles: Indexes = None) -> None:
        super().__init__('polyhedron',
                         {'points': points, 'faces': faces,
                          'convexity': convexity,
                          'triangles': triangles})


class union(OpenSCADObject):
    """
    Creates a union of all its child nodes. This is the **sum** of all
    children.
    """

    def __init__(self) -> None:
        super().__init__('union', {})

    def __add__(self, x: OpenSCADObjectPlus) -> OpenSCADObject:
        new_union = union()
        for child in self.children:
            new_union.add(child)
        new_union.add(x)

        return new_union


class intersection(OpenSCADObject):
    """
    Creates the intersection of all child nodes. This keeps the
    **overlapping** portion
    """

    def __init__(self) -> None:
        super().__init__('intersection', {})

    def __mul__(self, x: OpenSCADObjectPlus) -> OpenSCADObject:
        new_int = intersection()
        for child in self.children:
            new_int.add(child)
        new_int.add(x)

        return new_int


class difference(OpenSCADObject):
    """
    Subtracts the 2nd (and all further) child nodes from the first one.
    """

    def __init__(self) -> None:
        super().__init__('difference', {})

    def __sub__(self, x: OpenSCADObjectPlus) -> OpenSCADObject:
        new_diff = difference()
        for child in self.children:
            new_diff.add(child)
        new_diff.add(x)

        return new_diff


class hole(OpenSCADObject):
    def __init__(self) -> None:
        super().__init__('hole', {})
        self.set_hole(is_hole=True)


class part(OpenSCADObject):
    def __init__(self) -> None:
        super().__init__('part', {})
        self.set_part_root(is_root=True)


class translate(OpenSCADObject):
    """
    Translates (moves) its child elements along the specified vector.

    :param v: X, Y and Z translation
    :type v: 3 value sequence
    """

    def __init__(self, v: P3 = None) -> None:
        super().__init__('translate', {'v': v})


class scale(OpenSCADObject):
    """
    Scales its child elements using the specified vector.

    :param v: X, Y and Z scale factor
    :type v: 3 value sequence
    """

    def __init__(self, v: P3 = None) -> None:
        super().__init__('scale', {'v': v})


class rotate(OpenSCADObject):
    """
    Rotates its child 'a' degrees about the origin of the coordinate system
    or around an arbitrary axis.

    :param a: degrees of rotation, or sequence for degrees of rotation in each of the X, Y and Z axis.
    :type a: number or 3 value sequence

    :param v: sequence specifying 0 or 1 to indicate which axis to rotate by 'a' degrees. Ignored if 'a' is a sequence.
    :type v: 3 value sequence
    """

    def __init__(self, a: Union[float, Vec3] = None, v: Vec3 = None) -> None:
        super().__init__('rotate', {'a': a, 'v': v})


class mirror(OpenSCADObject):
    """
    Mirrors the child element on a plane through the origin.

    :param v: the normal vector of a plane intersecting the origin through which to mirror the object.
    :type v: 3 number sequence

    """

    def __init__(self, v: Vec3) -> None:
        super().__init__('mirror', {'v': v})


class resize(OpenSCADObject):
    """
    Modify the size of the child object to match the given new size.

    :param newsize: X, Y and Z values
    :type newsize: 3 value sequence
    
    :param auto: 3-tuple of booleans to specify which axes should be scaled
    :type auto: 3 boolean sequence
    """

    def __init__(self, newsize: Vec3, auto: Tuple[bool, bool, bool] = None) -> None:
        super().__init__('resize', {'newsize': newsize, 'auto': auto})


class multmatrix(OpenSCADObject):
    """
    Multiplies the geometry of all child elements with the given 4x4
    transformation matrix.

    :param m: transformation matrix
    :type m: sequence of 4 sequences, each containing 4 numbers.
    """

    def __init__(self, m: Tuple[Vec4, Vec4, Vec4, Vec4]) -> None:
        super().__init__('multmatrix', {'m': m})


class color(OpenSCADObject):
    """
    Displays the child elements using the specified RGB color + alpha value.
    This is only used for the F5 preview as CGAL and STL (F6) do not
    currently support color. The alpha value will default to 1.0 (opaque) if
    not specified.

    :param c: RGB color + alpha value.
    :type c: sequence of 3 or 4 numbers between 0 and 1, OR 3-, 4-, 6-, or 8-digit RGB/A hex code, OR string color name as described at https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#color

    :param alpha: Alpha value from 0 to 1
    :type alpha: float 
    """

    def __init__(self, c: Union[Vec34, str], alpha: float = 1.0) -> None:
        super().__init__('color', {'c': c, 'alpha': alpha})


class minkowski(OpenSCADObject):
    """
    Renders the `minkowski
    sum <http://www.cgal.org/Manual/latest/doc_html/cgal_manual/Minkowski_sum_3/Chapter_main.html>`__
    of child nodes.
    """

    def __init__(self) -> None:
        super().__init__('minkowski', {})


class offset(OpenSCADObject):
    """
    
    :param r: Amount to offset the polygon (rounded corners). When negative, 
        the polygon is offset inwards. The parameter r specifies the radius 
        that is used to generate rounded corners, using delta gives straight edges.
    :type r: number
    
    :param delta: Amount to offset the polygon (sharp corners). When negative, 
        the polygon is offset inwards. The parameter r specifies the radius 
        that is used to generate rounded corners, using delta gives straight edges.
    :type delta: number
    
    :param chamfer: When using the delta parameter, this flag defines if edges 
        should be chamfered (cut off with a straight line) or not (extended to 
        their intersection).
    :type chamfer: bool

    :param segments: Resolution of any radial curves
    :type segments: int
    """

    def __init__(self, r: float = None, delta: float = None, chamfer: bool = False,
                segments: int=None) -> None:
        if r is not None:
            kwargs = {'r': r}
        elif delta is not None:
            kwargs = {'delta': delta, 'chamfer': chamfer}
        else:
            raise ValueError("offset(): Must supply r or delta")
        if segments:
            kwargs['segments'] = segments
        super().__init__('offset', kwargs)


class hull(OpenSCADObject):
    """
    Renders the `convex
    hull <http://www.cgal.org/Manual/latest/doc_html/cgal_manual/Convex_hull_2/Chapter_main.html>`__
    of child nodes.
    """

    def __init__(self) -> None:
        super().__init__('hull', {})


class render(OpenSCADObject):
    """
    Always calculate the CSG model for this tree (even in OpenCSG preview
    mode).

    :param convexity: The convexity parameter specifies the maximum number of front sides (back sides) a ray intersecting the object might penetrate. This parameter is only needed for correctly displaying the object in OpenCSG preview mode and has no effect on the polyhedron rendering.
    :type convexity: int
    """

    def __init__(self, convexity: int = None) -> None:
        super().__init__('render', {'convexity': convexity})


class linear_extrude(OpenSCADObject):
    """
    Linear Extrusion is a modeling operation that takes a 2D polygon as
    input and extends it in the third dimension. This way a 3D shape is
    created.

    :param height: the extrusion height.
    :type height: number

    :param center: determines if the object is centered on the Z-axis after extrusion.
    :type center: boolean

    :param convexity: The convexity parameter specifies the maximum number of 
    front sides (back sides) a ray intersecting the object might penetrate. This 
    parameter is only needed for correctly displaying the object in OpenCSG 
    preview mode and has no effect on the polyhedron rendering.
    :type convexity: int

    :param twist: Twist is the number of degrees of through which the shape is 
    extruded.  Setting to 360 will extrude through one revolution.  The twist 
    direction follows the left hand rule.
    :type twist: number

    :param slices: number of slices to extrude. Can be used to improve the output.
    :type slices: int

    :param scale: relative size of the top of the extrusion compared to the start
    :type scale: number

    """

    def __init__(self, height: float = None, center: bool = None, convexity: int = None,
                 twist: float = None, slices: int = None, scale: float = None) -> None:
        super().__init__('linear_extrude',
                         {'height': height, 'center': center,
                          'convexity': convexity, 'twist': twist,
                          'slices': slices, 'scale': scale})


class rotate_extrude(OpenSCADObject):
    """
    A rotational extrusion is a Linear Extrusion with a twist, literally.
    Unfortunately, it can not be used to produce a helix for screw threads
    as the 2D outline must be normal to the axis of rotation, ie they need
    to be flat in 2D space.

    The 2D shape needs to be either completely on the positive, or negative
    side (not recommended), of the X axis. It can touch the axis, i.e. zero,
    however if the shape crosses the X axis a warning will be shown in the
    console windows and the rotate/_extrude() will be ignored. If the shape
    is in the negative axis the faces will be inside-out, you probably don't
    want to do that; it may be fixed in the future.

    :param angle: Defaults to 360. Specifies the number of degrees to sweep, 
    starting at the positive X axis. The direction of the sweep follows the 
    Right Hand Rule, hence a negative angle will sweep clockwise.
    :type angle: number
    
    :param segments: Number of fragments in 360 degrees.
    :type segments: int

    :param convexity: The convexity parameter specifies the maximum number of 
    front sides (back sides) a ray intersecting the object might penetrate. This 
    parameter is only needed for correctly displaying the object in OpenCSG 
    preview mode and has no effect on the polyhedron rendering.
    :type convexity: int

    """

    def __init__(self, angle: float = 360, convexity: int = None, segments: int = None) -> None:
        super().__init__('rotate_extrude',
                         {'angle': angle, 'segments': segments,
                          'convexity': convexity})


class dxf_linear_extrude(OpenSCADObject):
    def __init__(self, file: PathStr, layer: float = None, height: float = None,
                 center: bool = None, convexity: int = None, twist: float = None,
                 slices: int = None) -> None:
        super().__init__('dxf_linear_extrude',
                         {'file': Path(file).as_posix(), 'layer': layer,
                          'height': height, 'center': center,
                          'convexity': convexity, 'twist': twist,
                          'slices': slices})


class projection(OpenSCADObject):
    """
    Creates 2d shapes from 3d models, and export them to the dxf format.
    It works by projecting a 3D model to the (x,y) plane, with z at 0.

    :param cut: when True only points with z=0 will be considered (effectively 
    cutting the object) When False points above and below the plane will be 
    considered as well (creating a proper projection).
    :type cut: boolean
    """

    def __init__(self, cut: bool = None) -> None:
        super().__init__('projection', {'cut': cut})


class surface(OpenSCADObject):
    """
    Surface reads information from text or image files.

    :param file: The path to the file containing the heightmap data.
    :type file: PathStr

    :param center: This determines the positioning of the generated object. If 
    True, object is centered in X- and Y-axis. Otherwise, the object is placed 
    in the positive quadrant. Defaults to False.
    :type center: boolean

    :param invert: Inverts how the color values of imported images are translated 
    into height values. This has no effect when importing text data files. 
    Defaults to False.
    :type invert: boolean

    :param convexity: The convexity parameter specifies the maximum number of 
    front sides (back sides) a ray intersecting the object might penetrate. 
    This parameter is only needed for correctly displaying the object in OpenCSG 
    preview mode and has no effect on the polyhedron rendering.
    :type convexity: int
    """

    def __init__(self, file, center: bool = None, convexity: int = None, invert=None) -> None:
        super().__init__('surface',
                         {'file': file, 'center': center,
                          'convexity': convexity, 'invert': invert})


class text(OpenSCADObject):
    """
    Create text using fonts installed on the local system or provided as separate 
    font file.

    :param text: The text to generate.
    :type text: string

    :param size: The generated text will have approximately an ascent of the given 
    value (height above the baseline). Default is 10.  Note that specific fonts 
    will vary somewhat and may not fill the size specified exactly, usually 
    slightly smaller.
    :type size: number

    :param font: The name of the font that should be used. This is not the name 
    of the font file, but the logical font name (internally handled by the 
    fontconfig library). A list of installed fonts can be obtained using the 
    font list dialog (Help -> Font List).
    :type font: string

    :param halign: The horizontal alignment for the text. Possible values are 
    "left", "center" and "right". Default is "left".
    :type halign: string

    :param valign: The vertical alignment for the text. Possible values are 
    "top", "center", "baseline" and "bottom". Default is "baseline".
    :type valign: string

    :param spacing: Factor to increase/decrease the character spacing.  The 
    default value of 1 will result in the normal spacing for the font, giving 
    a value greater than 1 will cause the letters to be spaced further apart.
    :type spacing: number

    :param direction: Direction of the text flow. Possible values are "ltr" 
    (left-to-right), "rtl" (right-to-left), "ttb" (top-to-bottom) and "btt" 
    (bottom-to-top). Default is "ltr".
    :type direction: string

    :param language: The language of the text. Default is "en".
    :type language: string

    :param script: The script of the text. Default is "latin".
    :type script: string

    :param segments: used for subdividing the curved path segments provided by 
    freetype
    :type segments: int
    """

    def __init__(self, text: str, size: float = None, font: str = None, halign: str = None,
                 valign: str = None, spacing: float = None, direction: str = None,
                 language: str = None, script: str = None, segments: int = None) -> None:
        super().__init__('text',
                         {'text': text, 'size': size, 'font': font,
                          'halign': halign, 'valign': valign,
                          'spacing': spacing, 'direction': direction,
                          'language': language, 'script': script,
                          'segments': segments})


class child(OpenSCADObject):
    def __init__(self, index: int = None, vector: Sequence[int] = None, range=None) -> None:
        super().__init__('child',
                         {'index': index, 'vector': vector,
                          'range': range})


class children(OpenSCADObject):
    """
    The child nodes of the module instantiation can be accessed using the
    children() statement within the module. The number of module children
    can be accessed using the $children variable.

    :param index: select one child, at index value. Index start at 0 and should 
    be less than or equal to $children-1.
    :type index: int

    :param vector: select children with index in vector. Index should be between 
    0 and $children-1.
    :type vector: sequence of int

    :param range: [:] or [::]. select children between to , incremented by (default 1).
    """

    def __init__(self, index: int = None, vector: float = None, range: P23 = None) -> None:
        super().__init__('children',
                         {'index': index, 'vector': vector,
                          'range': range})


class import_stl(OpenSCADObject):
    def __init__(self, file: PathStr, origin: P2 = (0, 0), convexity: int = None, layer: int = None) -> None:
        super().__init__('import',
                         {'file': Path(file).as_posix(), 'origin': origin,
                          'convexity': convexity, 'layer': layer})


class import_dxf(OpenSCADObject):
    def __init__(self, file, origin=(0, 0), convexity: int = None, layer: int = None) -> None:
        super().__init__('import',
                         {'file': file, 'origin': origin,
                          'convexity': convexity, 'layer': layer})


class import_(OpenSCADObject):
    """
    Imports a file for use in the current OpenSCAD model. OpenSCAD currently
    supports import of DXF and STL (both ASCII and Binary) files.

    :param file: path to the STL or DXF file.
    :type file: PathStr

    :param convexity: The convexity parameter specifies the maximum number of 
    front sides (back sides) a ray intersecting the object might penetrate. This 
    parameter is only needed for correctly displaying the object in OpenCSG 
    preview mode and has no effect on the polyhedron rendering.
    :type convexity: int
    """

    def __init__(self, file: PathStr, origin: P2 = (0, 0), convexity: int = None, layer: int = None) -> None:
        super().__init__('import',
                         {'file': Path(file).as_posix(), 'origin': origin,
                          'convexity': convexity, 'layer': layer})


class intersection_for(OpenSCADObject):
    """
    Iterate over the values in a vector or range and take an
    intersection of the contents.
    """

    def __init__(self, n: int) -> None:
        super().__init__('intersection_for', {'n': n})


class assign(OpenSCADObject):
    def __init__(self) -> None:
        super().__init__('assign', {})


# ================================
# = Modifier Convenience Methods =
# ================================
def debug(openscad_obj: OpenSCADObject) -> OpenSCADObject:
    openscad_obj.set_modifier("#")
    return openscad_obj


def background(openscad_obj: OpenSCADObject) -> OpenSCADObject:
    openscad_obj.set_modifier("%")
    return openscad_obj


def root(openscad_obj: OpenSCADObject) -> OpenSCADObject:
    openscad_obj.set_modifier("!")
    return openscad_obj


def disable(openscad_obj: OpenSCADObject) -> OpenSCADObject:
    openscad_obj.set_modifier("*")
    return openscad_obj


# ===========================
# = IMPORTING OPENSCAD CODE =
# ===========================
def import_scad(scad_file_or_dir: PathStr) -> SimpleNamespace:
    '''
    Recursively look in current directory & OpenSCAD library directories for
        OpenSCAD files. Create Python mappings for all OpenSCAD modules & functions
    Return a namespace or raise ValueError if no scad files found
    '''
    global IMPORTED_SCAD_MODULES
    
    scad = Path(scad_file_or_dir)
    candidates: List[Path] = [scad]

    ns = IMPORTED_SCAD_MODULES.get(scad)
    if ns:
        return ns
    else:
        if not scad.is_absolute():
            candidates = [d/scad for d in _openscad_library_paths()]

        for candidate_path in candidates:
            namespace = _import_scad(candidate_path)
            if namespace is not None:
                IMPORTED_SCAD_MODULES[scad] = namespace
                return namespace
    raise ValueError(f'Could not find .scad files at or under {scad}. \nLocations searched were: {candidates}')

def _import_scad(scad: Path) -> Optional[SimpleNamespace]:
    '''
    cases:
        single scad file:
            return a namespace populated with `use()`
        directory
            recurse into all subdirectories and *.scad files
            return namespace if scad files are underneath, otherwise None
        non-scad file:
            return None            
    '''
    namespace: Optional[SimpleNamespace] = None
    if scad.is_file() and scad.suffix == '.scad':
        namespace = SimpleNamespace()
        use(scad.absolute(), dest_namespace_dict=namespace.__dict__)
    elif scad.is_dir():
        subspaces = [(f, _import_scad(f)) for f in scad.iterdir() if f.is_dir() or f.suffix == '.scad']
        for f, subspace in subspaces:
            if subspace:
                if namespace is None:
                    namespace = SimpleNamespace()
                # Add a subspace to namespace named by the file/dir it represents
                package_name = f.stem
                # Prefix an underscore to packages starting with a digit, which
                # are valid in OpenSCAD but not in Python
                if package_name[0].isdigit():
                    package_name = '_' + package_name

                setattr(namespace, package_name, subspace)

    return namespace
   
def _openscad_library_paths() -> List[Path]:
    """
    Return system-dependent OpenSCAD library paths or paths defined in os.environ['OPENSCADPATH']
    """
    import platform
    import os
    import re

    paths = [Path('.')]

    user_path = os.environ.get('OPENSCADPATH')
    if user_path:
        for s in re.split(r'\s*[;:]\s*', user_path):
            paths.append(Path(s))

    default_paths = {
        'Linux':   [
            Path.home() / '.local/share/OpenSCAD/libraries',
            Path('/usr/share/openscad/libraries')
        ],
        'Darwin':  [
            Path.home() / 'Documents/OpenSCAD/libraries',
            Path('/Applications/OpenSCAD.app/Contents/Resources/libraries')
        ],
        'Windows': [
            Path('My Documents\OpenSCAD\libraries'),
            Path('c:\Program Files\OpenSCAD\libraries')
        ],
    }

    paths += default_paths.get(platform.system(), [])
    return paths


def _find_library(library_name: PathStr) -> Path:
    result = Path(library_name)

    if not result.is_absolute():
        paths = _openscad_library_paths()
        for p in paths:
            f = p / result
            # print(f'Checking {f} -> {f.exists()}')
            if f.exists():
                result = f
                break

    return result

# use() & include() mimic OpenSCAD's use/include mechanics.
# -- use() makes methods in scad_file_path.scad available to be called.
# --include() makes those methods available AND executes all code in
#   scad_file_path.scad, which may have side effects.
#   Unless you have a specific need, call use().


def use(scad_file_path: PathStr, use_not_include: bool = True, dest_namespace_dict: Dict = None):
    """
    Opens scad_file_path, parses it for all usable calls,
    and adds them to caller's namespace.
    """
    # These functions in solidpython are used here and only here; don't pollute
    # the global namespace with them
    from .solidpython import parse_scad_callables
    from .solidpython import new_openscad_class_str
    from .solidpython import calling_module

    scad_file_path = _find_library(scad_file_path) 

    symbols_dicts = parse_scad_callables(scad_file_path)

    for sd in symbols_dicts:
        class_str = new_openscad_class_str(sd['name'], sd['args'], sd['kwargs'],
                                           scad_file_path.as_posix(), use_not_include)
        # If this is called from 'include', we have to look deeper in the stack
        # to find the right module to add the new class to.
        if dest_namespace_dict is None:
            stack_depth = 2 if use_not_include else 3
            dest_namespace_dict = calling_module(stack_depth).__dict__
        try:
            exec(class_str, dest_namespace_dict)
        except Exception as e:
            classname = sd['name']
            msg = f"Unable to import SCAD module: `{classname}` from `{scad_file_path.name}`, with error: {e}"
            print(msg)

    return True

def include(scad_file_path: PathStr) -> bool:
    return use(scad_file_path, use_not_include=False)




"""
A version of unittest that gives output in an easier to use format
"""
import sys
import unittest
import difflib


class DiffOutput(unittest.TestCase):

    def assertEqual(self, first, second, msg=None):
        """
        Override assertEqual and print(a context diff if msg=None)
        """
        # Test if both are strings, in Python 2 & 3
        string_types = str if sys.version_info[0] == 3 else basestring

        if isinstance(first, string_types) and isinstance(second, string_types):
            if not msg:
                msg = 'Strings are not equal:\n' + ''.join(
                    difflib.unified_diff(
                        [first],
                        [second],
                        fromfile='actual',
                        tofile='expected'
                    )
                )
        return super(DiffOutput, self).assertEqual(first, second, msg=msg)



#! /usr/bin/env python

#    Simple Python OpenSCAD Code Generator
#    Copyright (C) 2009    Philipp Tiefenbacher <wizards23@gmail.com>
#    Amendments & additions, (C) 2011-2019 Evan Jones <evan_t_jones@mac.com>
#
#   License: LGPL 2.1 or later
#

from __future__ import annotations

import datetime
import inspect
import os
import subprocess
import sys
import tempfile
from pathlib import Path
import keyword

from typing import Set, Sequence, List, Callable, Optional, Union, Iterable

from types import ModuleType
from typing import Callable, Iterable, List, Optional, Sequence, Set, Union, Dict

import pkg_resources
import re

PathStr = Union[Path, str]
AnimFunc = Callable[[Optional[float]], 'OpenSCADObject']
# These are features added to SolidPython but NOT in OpenSCAD.
# Mark them for special treatment
non_rendered_classes = ['hole', 'part']

# Words reserved in Python but not OpenSCAD
# Re: https://github.com/SolidCode/SolidPython/issues/99

PYTHON_ONLY_RESERVED_WORDS = keyword.kwlist


# =========================
# = Internal Utilities    =
# =========================
class OpenSCADObject:

    def __init__(self, name: str, params: dict):
        self.name = name
        self.params = params
        self.children: List["OpenSCADObject"] = []
        self.modifier = ""
        self.parent: Optional["OpenSCADObject"] = None
        self.is_hole = False
        self.has_hole_children = False
        self.is_part_root = False
        self.traits: Dict[str, Dict[str, float]] = {}

    def add_trait(self, trait_name:str, trait_data:Dict[str, float]):
        self.traits[trait_name] = trait_data

    def get_trait(self, trait_name:str) -> Optional[Dict[str, float]]:
        return self.traits.get(trait_name)

    def set_hole(self, is_hole: bool = True) -> "OpenSCADObject":
        self.is_hole = is_hole
        return self

    def set_part_root(self, is_root: bool = True) -> "OpenSCADObject":
        self.is_part_root = is_root
        return self

    def find_hole_children(self, path: List["OpenSCADObject"] = None) -> List["OpenSCADObject"]:
        """
        Because we don't force a copy every time we re-use a node
        (e.g a = cylinder(2, 6);  b = right(10) (a)
         the identical 'a' object appears in the tree twice),
        we can't count on an object's 'parent' field to trace its
        path to the root.  Instead, keep track explicitly
        """
        path = path if path else [self]
        hole_kids = []

        for child in self.children:
            path.append(child)
            if child.is_hole:
                hole_kids.append(child)
                # Mark all parents as having a hole child
                for p in path:
                    p.has_hole_children = True
            # Don't append holes from separate parts below us
            elif child.is_part_root:
                continue
            # Otherwise, look below us for children
            else:
                hole_kids += child.find_hole_children(path)
            path.pop()

        return hole_kids

    def set_modifier(self, m: str) -> "OpenSCADObject":
        """
        Used to add one of the 4 single-character modifiers:
        #(debug) !(root) %(background) or *(disable)
        """
        string_vals = {'disable': '*',
                       'debug': '#',
                       'background': '%',
                       'root': '!',
                       '*': '*',
                       '#': '#',
                       '%': '%',
                       '!': '!'}

        self.modifier = string_vals.get(m.lower(), '')
        return self

    def _render(self, render_holes: bool = False) -> str:
        """
        NOTE: In general, you won't want to call this method. For most purposes,
        you really want scad_render(), 
        Calling obj._render won't include necessary 'use' or 'include' statements
        """
        # First, render all children
        s = ""
        for child in self.children:
            # Don't immediately render hole children.
            # Add them to the parent's hole list,
            # And render after everything else
            if not render_holes and child.is_hole:
                continue
            s += child._render(render_holes)

        # Then render self and prepend/wrap it around the children
        # I've added designated parts and explicit holes to SolidPython.
        # OpenSCAD has neither, so don't render anything from these objects
        if self.name in non_rendered_classes:
            pass
        elif not self.children:
            s = self._render_str_no_children() + ";"
        else:
            s = self._render_str_no_children() + " {" + indent(s) + "\n}"

        # If this is the root object or the top of a separate part,
        # find all holes and subtract them after all positive geometry
        # is rendered
        if (not self.parent) or self.is_part_root:
            hole_children = self.find_hole_children()

            if len(hole_children) > 0:
                s += "\n/* Holes Below*/"
                s += self._render_hole_children()

                # wrap everything in the difference
                s = "\ndifference(){" + indent(s) + " /* End Holes */ \n}"
        return s

    def _render_str_no_children(self) -> str:
        callable_name = _unsubbed_keyword(self.name)
        s = "\n" + self.modifier + callable_name + "("
        first = True

        # Re: https://github.com/SolidCode/SolidPython/issues/99
        # OpenSCAD will accept Python reserved words as callables or argument names,
        # but they won't compile in Python. Those have already been substituted
        # out (e.g 'or' => 'or_'). Sub them back here.
        self.params = {_unsubbed_keyword(k): v for k, v in self.params.items()}

        # OpenSCAD doesn't have a 'segments' argument, but it does
        # have '$fn'.  Swap one for the other
        if 'segments' in self.params:
            self.params['$fn'] = self.params.pop('segments')

        valid_keys = self.params.keys()

        # intkeys are the positional parameters
        intkeys = list(filter(lambda x: type(x) == int, valid_keys))
        intkeys.sort()

        # named parameters
        nonintkeys = list(filter(lambda x: not type(x) == int, valid_keys))
        all_params_sorted = intkeys + nonintkeys
        if all_params_sorted:
            all_params_sorted = sorted(all_params_sorted)

        for k in all_params_sorted:
            v = self.params[k]
            if v is None:
                continue

            if not first:
                s += ", "
            first = False

            if type(k) == int:
                s += py2openscad(v)
            else:
                s += k + " = " + py2openscad(v)

        s += ")"
        return s

    def _render_hole_children(self) -> str:
        # Run down the tree, rendering only those nodes
        # that are holes or have holes beneath them
        if not self.has_hole_children:
            return ""
        s = ""
        for child in self.children:
            if child.is_hole:
                s += child._render(render_holes=True)
            elif child.has_hole_children:
                s += child._render_hole_children()
        if self.name in non_rendered_classes:
            pass
        else:
            s = self._render_str_no_children() + "{" + indent(s) + "\n}"

        # Holes exist in the compiled tree in two pieces:
        # The shapes of the holes themselves, (an object for which
        # obj.is_hole is True, and all its children) and the
        # transforms necessary to put that hole in place, which
        # are inherited from non-hole geometry.

        # Non-hole Intersections & differences can change (shrink)
        # the size of holes, and that shouldn't happen: an
        # intersection/difference with an empty space should be the
        # entirety of the empty space.
        #  In fact, the intersection of two empty spaces should be
        # everything contained in both of them:  their union.
        # So... replace all super-hole intersection/diff transforms
        # with union in the hole segment of the compiled tree.
        # And if you figure out a better way to explain this,
        # please, please do... because I think this works, but I
        # also think my rationale is shaky and imprecise. 
        # -ETJ 19 Feb 2013
        s = s.replace("intersection", "union")
        s = s.replace("difference", "union")

        return s

    def add(self, child: Union["OpenSCADObject", Sequence["OpenSCADObject"]]) -> "OpenSCADObject":
        """
        if child is a single object, assume it's an OpenSCADObjects and 
        add it to self.children

        if child is a list, assume its members are all OpenSCADObjects and
        add them all to self.children
        """
        if isinstance(child, (list, tuple)):
            # __call__ passes us a list inside a tuple, but we only care
            # about the list, so skip single-member tuples containing lists
            if len(child) == 1 and isinstance(child[0], (list, tuple)):
                child = child[0]
            [self.add(c) for c in child]
        elif isinstance(child, int):
            # Allowing for creating object by adding to 0 (as in sum())
            if child != 0:
                raise ValueError
        else:
            self.children.append(child)  # type: ignore
            child.set_parent(self)  # type: ignore
        return self

    def set_parent(self, parent: "OpenSCADObject"):
        self.parent = parent

    def add_param(self, k: str, v: float) -> "OpenSCADObject":
        if k == '$fn':
            k = 'segments'
        self.params[k] = v
        return self

    def copy(self) -> "OpenSCADObject":
        """
        Provides a copy of this object and all children,
        but doesn't copy self.parent, meaning the new object belongs
        to a different tree
        Initialize an instance of this class with the same params
        that created self, the object being copied.
        """

        # Python can't handle an '$fn' argument, while openSCAD only wants
        # '$fn'.  Swap back and forth as needed; the final renderer will
        # sort this out.
        if '$fn' in self.params:
            self.params['segments'] = self.params.pop('$fn')

        other = type(self)(**self.params)
        other.set_modifier(self.modifier)
        other.set_hole(self.is_hole)
        other.set_part_root(self.is_part_root)
        other.has_hole_children = self.has_hole_children
        for c in self.children:
            other.add(c.copy())
        return other

    def __call__(self, *args: "OpenSCADObject") -> "OpenSCADObject":
        """
        Adds all objects in args to self.  This enables OpenSCAD-like syntax,
        e.g.:
        union()(
            cube(),
            sphere()
        )
        """
        return self.add(args)

    def __add__(self, x: "OpenSCADObject") -> "OpenSCADObject":
        """
        This makes u = a+b identical to:
        u = union()(a, b )
        """
        return objects.union()(self, x)

    def __radd__(self, x: "OpenSCADObject") -> "OpenSCADObject":
        """
        This makes u = a+b identical to:
        u = union()(a, b )
        """
        return objects.union()(self, x)

    def __sub__(self, x: "OpenSCADObject") -> "OpenSCADObject":
        """
        This makes u = a - b identical to:
        u = difference()(a, b )
        """
        return objects.difference()(self, x)

    def __mul__(self, x: "OpenSCADObject") -> "OpenSCADObject":
        """
        This makes u = a * b identical to:
        u = intersection()(a, b )
        """
        return objects.intersection()(self, x)

    def _repr_png_(self) -> Optional[bytes]:
        """
        Allow rich clients such as the IPython Notebook, to display the current
        OpenSCAD rendering of this object.
        """
        png_data = None
        tmp = tempfile.NamedTemporaryFile(suffix=".scad", delete=False)
        tmp_png = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
        try:
            scad_text = scad_render(self).encode("utf-8")
            tmp.write(scad_text)
            tmp.close()
            tmp_png.close()
            subprocess.Popen([
                "openscad",
                "--preview",
                "-o", tmp_png.name,
                tmp.name
            ]).communicate()

            with open(tmp_png.name, "rb") as png:
                png_data = png.read()
        finally:
            os.unlink(tmp.name)
            os.unlink(tmp_png.name)

        return png_data


class IncludedOpenSCADObject(OpenSCADObject):
    """
    Identical to OpenSCADObject, but each subclass of IncludedOpenSCADObject
    represents imported scad code, so each instance needs to store the path
    to the scad file it's included from.
    """

    def __init__(self, name, params, include_file_path, use_not_include=False, **kwargs):
        self.include_file_path = self._get_include_path(include_file_path)

        use_str = 'use' if use_not_include else 'include'
        self.include_string = f'{use_str} <{self.include_file_path}>\n'

        # Just pass any extra arguments straight on to OpenSCAD; it'll accept
        # them
        if kwargs:
            params.update(kwargs)

        OpenSCADObject.__init__(self, name, params)

    def _get_include_path(self, include_file_path):
        # Look through sys.path for anyplace we can find a valid file ending
        # in include_file_path.  Return that absolute path
        if os.path.isabs(include_file_path) and os.path.isfile(include_file_path):
            return include_file_path
        else:
            for p in sys.path:
                whole_path = os.path.join(p, include_file_path)
                if os.path.isfile(whole_path):
                    return os.path.abspath(whole_path)

        # No loadable SCAD file was found in sys.path.  Raise an error
        raise ValueError(f"Unable to find included SCAD file: {include_file_path} in sys.path")


# =========================================
# = Rendering Python code to OpenSCAD code=
# =========================================
def _find_include_strings(obj: Union[IncludedOpenSCADObject, OpenSCADObject]) -> Set[str]:
    include_strings = set()
    if isinstance(obj, IncludedOpenSCADObject):
        include_strings.add(obj.include_string)
    for child in obj.children:
        include_strings.update(_find_include_strings(child))
    # We also accept IncludedOpenSCADObject instances as parameters to functions, 
    # so search in obj.params as well
    for param in obj.params.values():
        if isinstance(param, OpenSCADObject):
            include_strings.update(_find_include_strings(param))
    return include_strings

def scad_render(scad_object: OpenSCADObject, file_header: str = '') -> str:
    # Make this object the root of the tree
    root = scad_object

    # Scan the tree for all instances of
    # IncludedOpenSCADObject, storing their strings
    include_strings = _find_include_strings(root)

    # and render the string
    includes = ''.join(include_strings) + "\n"
    scad_body = root._render()

    if file_header and not file_header.endswith('\n'): 
        file_header += '\n'

    return file_header + includes + scad_body

def scad_render_animated(func_to_animate: AnimFunc, 
                         steps: int =20, 
                         back_and_forth: bool=True, 
                         file_header: str='') -> str:
    # func_to_animate takes a single float argument, _time in [0, 1), and
    # returns an OpenSCADObject instance.
    #
    # Outputs an OpenSCAD file with func_to_animate() evaluated at "steps"
    # points between 0 & 1, with time never evaluated at exactly 1

    # If back_and_forth is True, smoothly animate the full extent of the motion
    # and then reverse it to the beginning; this avoids skipping between beginning
    # and end of the animated motion

    # NOTE: This is a hacky way to solve a simple problem.  To use OpenSCAD's
    # animation feature, our code needs to respond to changes in the value
    # of the OpenSCAD variable $t, but I can't think of a way to get a
    # float variable from our code and put it into the actual SCAD code.
    # Instead, we just evaluate our code at each desired step, and write it
    # all out in the SCAD code for each case, with an if/else tree.  Depending
    # on the number of steps, this could create hundreds of times more SCAD
    # code than is needed.  But... it does work, with minimal Python code, so
    # here it is. Better solutions welcome. -ETJ 28 Mar 2013

    # NOTE: information on the OpenSCAD manual wiki as of November 2012 implies
    # that the OpenSCAD app does its animation irregularly; sometimes it animates
    # one loop in steps iterations, and sometimes in (steps + 1).  Do it here
    # in steps iterations, meaning that we won't officially reach $t =1.

    # Note also that we check for ranges of time rather than equality; this
    # should avoid any rounding error problems, and doesn't require the file
    # to be animated with an identical number of steps to the way it was
    # created. -ETJ 28 Mar 2013
    scad_obj = func_to_animate(_time=0)  # type: ignore
    include_strings = _find_include_strings(scad_obj)
    # and render the string
    includes = ''.join(include_strings) + "\n"

    rendered_string = file_header + includes

    if back_and_forth:
        steps *= 2

    for i in range(steps):
        time = i * 1.0 / steps
        end_time = (i + 1) * 1.0 / steps
        eval_time = time
        # Looping back and forth means there's no jump between the start and
        # end position
        if back_and_forth:
            if time < 0.5:
                eval_time = time * 2
            else:
                eval_time = 2 - 2 * time
        scad_obj = func_to_animate(_time=eval_time)  # type: ignore

        scad_str = indent(scad_obj._render())
        rendered_string += f"if ($t >= {time} && $t < {end_time}){{" \
                           f"   {scad_str}\n" \
                           f"}}\n"
    return rendered_string

def scad_render_animated_file(func_to_animate:AnimFunc, 
                              steps: int=20, 
                              back_and_forth: bool=True, 
                              filepath: Optional[str]=None, 
                              out_dir: PathStr=None, 
                              file_header: str='', 
                              include_orig_code: bool=True) -> str:
    rendered_string = scad_render_animated(func_to_animate, steps, 
                                            back_and_forth, file_header)
    return _write_code_to_file(rendered_string, filepath, out_dir=out_dir, 
                include_orig_code=include_orig_code)

def scad_render_to_file(scad_object: OpenSCADObject,
                        filepath: PathStr=None, 
                        out_dir: PathStr=None,
                        file_header: str='', 
                        include_orig_code: bool=True) -> str:
    header = file_header
    if include_orig_code:
        version = _get_version()
        date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        header = f"// Generated by SolidPython {version} on {date}\n" + file_header

    rendered_string = scad_render(scad_object, header)
    return _write_code_to_file(rendered_string, filepath, out_dir, include_orig_code)

def _write_code_to_file(rendered_string: str, 
                        filepath: PathStr=None, 
                        out_dir: PathStr=None, 
                        include_orig_code: bool=True) -> str:
    try:
        calling_file = Path(calling_module(stack_depth=3).__file__).absolute()
        # Output path is determined four ways:
        # -- If filepath is supplied, use filepath
        # -- If no filepath is supplied but an out_dir is supplied, 
        #       give the calling file a .scad suffix and put it in out_dir
        # -- If neither filepath nor out_dir are supplied, give the new
        #       file a .scad suffix and put it next to the calling file
        # -- If no path info is supplied and we can't find a calling file 
        #       (i.e, this is being called from an interactive terminal), 
        #       write a file to Path.cwd() / 'solid.scad'
        out_path = Path()
        if filepath:
            out_path = Path(filepath)
        elif out_dir:
            odp = Path(out_dir)
            if not odp.exists():
                odp.mkdir()
            out_path = odp / calling_file.with_suffix('.scad').name
        else:
            out_path = calling_file.with_suffix('.scad')
        
        if include_orig_code:
            rendered_string += sp_code_in_scad_comment(calling_file)
    except AttributeError as e:
        # If no calling_file was found, this is being called from the terminal.
        # We can't read original code from a file, so don't try,
        # and can't read filename from the calling file either, so just save to
        # solid.scad.

        if filepath:
            out_path = Path(filepath)
        else:
            odp = Path(out_dir) if out_dir else Path.cwd()
            if not odp.exists():
                odp.mkdir()
            out_path = odp / 'solid.scad'

    out_path.write_text(rendered_string)
    return out_path.absolute().as_posix()

def _get_version() -> str:
    """
    Returns SolidPython version
    Returns '<Unknown>' if no version can be found
    """
    version = '<Unknown>'
    try:
        # if SolidPython is installed use `pkg_resources`
        version = pkg_resources.get_distribution('solidpython').version

    except pkg_resources.DistributionNotFound:
        # if the running SolidPython is not the one installed via pip,
        # try to read it from the project setup file
        version_pattern = re.compile(r"version = ['\"]([^'\"]*)['\"]")
        version_file_path = Path(__file__).parent.parent / 'pyproject.toml'
        if version_file_path.exists():
            version_match = version_pattern.search(version_file_path.read_text())
            if version_match:
                version = version_match.group(1)
    return version

def sp_code_in_scad_comment(calling_file: PathStr) -> str:
    """
    Once a SCAD file has been created, it's difficult to reconstruct
    how it got there, since it has no variables, modules, etc.  So, include
    the Python code that generated the scad code as comments at the end of
    the SCAD code
    """
    pyopenscad_str = Path(calling_file).read_text()

    # TODO: optimally, this would also include a version number and
    # git hash (& date & github URL?) for the version of solidpython used
    # to create a given file; That would future-proof any given SP-created
    # code because it would point to the relevant dependencies as well as
    # the actual code
    pyopenscad_str = (f"\n"
                      f"/***********************************************\n"
                      f"*********      SolidPython code:      **********\n"
                      f"************************************************\n"
                      f" \n"
                      f"{pyopenscad_str} \n"
                      f" \n"
                      f"************************************************/\n")
    return pyopenscad_str


# ===========
# = Parsing =
# ===========
def parse_scad_callables(filename: str) -> List[dict]:
    from .py_scadparser import scad_parser

    modules, functions, _ = scad_parser.parseFile(filename)

    callables = []
    for c in modules + functions:
        args = []
        kwargs = []

        #for some reason solidpython needs to treat all openscad arguments as if
        #they where optional. I don't know why, but at least to pass the tests
        #it's neccessary to handle it like this !?!?!
        for p in c.parameters:
            kwargs.append(p.name)
            #if p.optional:
            #    kwargs.append(p.name)
            #else:
            #    args.append(p.name)

        callables.append({'name': c.name, 'args': args, 'kwargs': kwargs})

    return callables

def calling_module(stack_depth: int = 2) -> ModuleType:
    """
    Returns the module *2* back in the frame stack.  That means:
    code in module A calls code in module B, which asks calling_module()
    for module A.

    This means that we have to know exactly how far back in the stack
    our desired module is; if code in module B calls another function in 
    module B, we have to increase the stack_depth argument to account for
    this.

    Got that?
    """
    frm = inspect.stack()[stack_depth]
    calling_mod = inspect.getmodule(frm[0])
    # If calling_mod is None, this is being called from an interactive session.
    # Return that module.  (Note that __main__ doesn't have a __file__ attr,
    # but that's caught elsewhere.)
    if not calling_mod:
        import __main__ as calling_mod  # type: ignore
    return calling_mod

def new_openscad_class_str(class_name: str,
                           args: Sequence[str] = None,
                           kwargs: Sequence[str] = None,
                           include_file_path: Optional[str] = None,
                           use_not_include: bool = True) -> str:
    args_str = ''
    args_pairs = ''

    args = args or []
    kwargs = kwargs or []

    # Re: https://github.com/SolidCode/SolidPython/issues/99
    # Don't allow any reserved words as argument names or module names
    # (They might be valid OpenSCAD argument names, but not in Python)
    class_name = _subbed_keyword(class_name)

    args = map(_subbed_keyword, args)  # type: ignore
    for arg in args:
        args_str += ', ' + arg
        args_pairs += f"'{arg}':{arg}, "

    # kwargs have a default value defined in their SCAD versions.  We don't
    # care what that default value will be (SCAD will take care of that), just
    # that one is defined.
    kwargs = map(_subbed_keyword, kwargs)  # type: ignore
    for kwarg in kwargs:
        args_str += f', {kwarg}=None'
        args_pairs += f"'{kwarg}':{kwarg}, "

    if include_file_path:
        # include_file_path may include backslashes on Windows; escape them
        # again here so any backslashes don't get used as escape characters
        # themselves
        include_file_str = Path(include_file_path).as_posix()

        # NOTE the explicit import of 'solid' below. This is a fix for:
        # https://github.com/SolidCode/SolidPython/issues/20 -ETJ 16 Jan 2014
        result = (f"import solid\n"
                  f"class {class_name}(solid.IncludedOpenSCADObject):\n"
                  f"   def __init__(self{args_str}, **kwargs):\n"
                  f"       solid.IncludedOpenSCADObject.__init__(self, '{class_name}', {{{args_pairs} }}, include_file_path='{include_file_str}', use_not_include={use_not_include}, **kwargs )\n"
                  f"   \n"
                  f"\n")
    else:
        result = (f"class {class_name}(OpenSCADObject):\n"
                  f"   def __init__(self{args_str}):\n"
                  f"       OpenSCADObject.__init__(self, '{class_name}', {{{args_pairs }}})\n"
                  f"   \n"
                  f"\n")

    return result

def _subbed_keyword(keyword: str) -> str:
    """
    Append an underscore to any python reserved word.
    Prepend an underscore to any OpenSCAD identifier starting with a digit.
    No-op for all other strings, e.g. 'or' => 'or_', 'other' => 'other'
    """
    new_key = keyword

    if keyword in PYTHON_ONLY_RESERVED_WORDS:
        new_key = keyword + "_"

    elif keyword[0].isdigit():
        new_key = "_" + keyword

    elif keyword == "$fn":
        new_key = "segments"

    elif keyword[0] == "$":
        new_key = "__" + keyword[1:]

    if new_key != keyword:
        print(f"\nFound OpenSCAD code that's not compatible with Python. \n"
              f"Imported OpenSCAD code using `{keyword}` \n"
              f"can be accessed with `{new_key}` in SolidPython\n")
    return new_key

def _unsubbed_keyword(subbed_keyword: str) -> str:
    """
    Remove trailing underscore for already-subbed python reserved words.
    Remove prepending underscore if remaining identifier starts with a digit.
    No-op for all other strings: e.g. 'or_' => 'or', 'other_' => 'other_'
    """
    if subbed_keyword.endswith("_") and subbed_keyword[:-1] in PYTHON_ONLY_RESERVED_WORDS:
        return subbed_keyword[:-1]

    elif subbed_keyword.startswith("__"):
        return "$" + subbed_keyword[2:]

    elif subbed_keyword.startswith("_") and subbed_keyword[1].isdigit():
        return subbed_keyword[1:]

    elif subbed_keyword == "segments":
        return "$fn"

    return subbed_keyword

# now that we have the base class defined, we can do a circular import
from . import objects

def py2openscad(o: Union[bool, float, str, Iterable]) -> str:
    if type(o) == bool:
        return str(o).lower()
    if type(o) == float:
        return f"{o:.10f}"  # type: ignore
    if type(o) == str:
        return f'\"{o}\"'  # type: ignore
    if type(o).__name__ == "ndarray":
        import numpy  # type: ignore
        return numpy.array2string(o, separator=",", threshold=1000000000)
    if isinstance(o, IncludedOpenSCADObject):
        return o._render()[1:-1]
    if hasattr(o, "__iter__"):
        s = "["
        first = True
        for i in o:  # type: ignore
            if not first:
                s += ", "
            first = False
            s += py2openscad(i)
        s += "]"
        return s
    return str(o)

def indent(s: str) -> str:
    return s.replace("\n", "\n\t")
